# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for mocking the GCS bucket for testing."""
from collections import defaultdict
import contextlib
import datetime
import os
import pathlib
import tempfile
from typing import Any
from unittest import mock

from google.cloud import exceptions

from . import gcs


@contextlib.contextmanager
def gcs_mock(directory: str | None = None):
  """A context for mocking reads/writes to the vulnerabilities GCS bucket.
  
  If `directory` is set, blobs will be read from/written to files in the
  directory, which will remain after the context exits.
  Otherwise, blobs will be written to a temporary directory, which is deleted
  when the context exits.
  """
  with (tempfile.TemporaryDirectory()
        if directory is None else contextlib.nullcontext(directory)) as db_dir:
    pathlib.Path(db_dir, gcs.VULN_PB_PATH).mkdir(parents=True, exist_ok=True)
    bucket = _MockBucket(db_dir)
    with mock.patch(
        'osv.gcs.get_osv_bucket',
        return_value=bucket), mock.patch('osv.pubsub.publish_failure'):
      yield db_dir


class _MockBlob:
  """Mock google.cloud.storage.Blob with only necessary methods for tests."""

  def __init__(self, path: str, generations: dict[str, int] = None):
    self._path = path
    self.custom_time: datetime.datetime | None = None
    # store a reference to all the blob generations
    self._generations = generations

  def upload_from_string(self,
                         data: str | bytes,
                         content_type: str | None = None,
                         if_generation_match: Any | None = None):
    """Implements google.cloud.storage.Blob.upload_from_string."""
    del content_type  # Can't do anything with this.

    if if_generation_match not in (None, self._generations[self._path]):
      raise exceptions.PreconditionFailed('Generation mismatch')

    if isinstance(data, str):
      data = data.encode()
    with open(self._path, 'wb') as f:
      f.write(data)
    self._generations[self._path] += 1

    # Use the file's modified time to store the CustomTime metadata.
    if self.custom_time is not None:
      ts = self.custom_time.timestamp()
      os.utime(self._path, (ts, ts))

  def download_as_bytes(self) -> bytes:
    """Implements google.cloud.storage.Blob.download_as_bytes."""
    try:
      with open(self._path, 'rb') as f:
        return f.read()
    except FileNotFoundError as exc:
      raise exceptions.NotFound(self._path) from exc


class _MockBucket:
  """Mock google.cloud.storage.Bucket with only necessary methods for tests."""

  def __init__(self, db_dir: str):
    self._db_dir = db_dir
    self._generations = defaultdict(int)
    # Create a generation for any pre-existing files
    for root, _, blobs in os.walk(self._db_dir):
      for blob_name in blobs:
        self._generations[os.path.join(root, blob_name)] = 1

  def blob(self, blob_name: str) -> _MockBlob:
    return _MockBlob(os.path.join(self._db_dir, blob_name), self._generations)

  def get_blob(self, blob_name: str) -> _MockBlob | None:
    path = os.path.join(self._db_dir, blob_name)
    if not os.path.exists(path):
      return None
    blob = _MockBlob(path, self._generations)
    ts = os.path.getmtime(path)
    blob.custom_time = datetime.datetime.fromtimestamp(ts, datetime.UTC)
    blob.generation = self._generations[path]
    return blob



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Request helpers"""

import requests
from requests.adapters import HTTPAdapter, Retry

from .cache import Cache

DEFAULT_BACKOFF_FACTOR = 5
DEFAULT_RETRY_TOTAL = 7
DEFAULT_REDIS_TTL_SECONDS = 6 * 60 * 60


class RequestError(Exception):
  """Exception raised by request helper when response is not 200"""
  response: requests.Response

  def __init__(self, response: requests.Response):
    self.response = response
    super().__init__(f'{response.status_code} Server Error: {response.reason} '
                     f'for url: {response.request.url}')


class RequestHelper:
  """Request helper that manages caching and automatic retries"""
  retry_total: int
  backoff_factor: int
  cache: Cache = None
  cache_ttl: int

  def __init__(self,
               cache=None,
               backoff_factor=DEFAULT_BACKOFF_FACTOR,
               retry_total=DEFAULT_RETRY_TOTAL,
               cache_ttl=DEFAULT_REDIS_TTL_SECONDS):
    self.backoff_factor = backoff_factor
    self.retry_total = retry_total
    self.cache_ttl = cache_ttl
    self.cache = cache

  def get(self, url):
    """Getter method.

    Retrieves the text at the URL, using the cached result if available.

    Args:
      url: URL to retrieve

    Returns:
      the text at the URL.

    Raises:
      RequestError on a non-200 HTTP response.

    """
    if self.cache:
      cached_result = self.cache.get(url)
      if cached_result:
        return cached_result

    with requests.session() as session:
      retries = Retry(
          backoff_factor=self.backoff_factor,
          total=self.retry_total,
      )
      session.mount('https://', HTTPAdapter(max_retries=retries))
      session.headers.update({'User-Agent': 'osv.dev'})
      response = session.get(url)

      if response.status_code != 200:
        raise RequestError(response)

      text_response = response.text
      if self.cache:
        self.cache.set(url, text_response, self.cache_ttl)

      return text_response



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helpers."""

import re

from .ecosystems_base import EnumerableEcosystem, OrderedEcosystem
from .alpine import Alpine, APK
from .bioconductor import Bioconductor
from .cran import CRAN
from .debian import Debian, DPKG
from .haskell import Hackage, GHC
from .maven import Maven
from .nuget import NuGet
from .packagist import Packagist
from .pub import Pub
from .pypi import PyPI
from .redhat import RPM
from .rubygems import RubyGems
from .semver_ecosystem_helper import SemverEcosystem
from .ubuntu import Ubuntu

_ecosystems = {
    'AlmaLinux': RPM,
    'Alpaquita': APK,
    'Alpine': Alpine,
    'BellSoft Hardened Containers': APK,
    'Bioconductor': Bioconductor,
    'Bitnami': SemverEcosystem,
    'Chainguard': APK,
    'CRAN': CRAN,
    'crates.io': SemverEcosystem,
    'Debian': Debian,
    'Echo': DPKG,
    'GHC': GHC,
    'Go': SemverEcosystem,
    'Hackage': Hackage,
    'Hex': SemverEcosystem,
    'Mageia': RPM,
    'Maven': Maven,
    'MinimOS': APK,
    'npm': SemverEcosystem,
    'NuGet': NuGet,
    'openEuler': RPM,
    'openSUSE': RPM,
    'Packagist': Packagist,
    'Pub': Pub,
    'PyPI': PyPI,
    'Red Hat': RPM,
    'Rocky Linux': RPM,
    'RubyGems': RubyGems,
    'SUSE': RPM,
    'SwiftURL': SemverEcosystem,
    'Ubuntu': Ubuntu,
    'Wolfi': APK,

    # Ecosystems known in the schema, but without implementations.
    # Must be kept in sync with osv-schema.
    'Android': None,
    'ConanCenter': None,
    'GIT': None,
    'GitHub Actions': None,
    'Linux': None,
    'OSS-Fuzz': None,
    'Photon OS': None,
}


def is_semver(ecosystem: str) -> bool:
  """Returns whether an ecosystem uses 'SEMVER' range types"""
  return isinstance(get(ecosystem), SemverEcosystem)


def is_known(ecosystem: str) -> bool:
  """Returns whether an ecosystem is known to OSV
  (even if ordering is not supported)."""
  name, _, _ = ecosystem.partition(':')
  return name in _ecosystems


package_urls = {
    'Android': 'https://android.googlesource.com/',
    'CRAN': 'https://cran.r-project.org/web/packages/',
    'crates.io': 'https://crates.io/crates/',
    'Debian': 'https://packages.debian.org/src:',
    'GitHub Actions': 'https://github.com/marketplace/actions/',
    'Go': 'https://',
    'Hackage': 'https://hackage.haskell.org/package/',
    'Hex': 'https://hex.pm/packages/',
    'Mageia': 'https://madb.mageia.org/show?rpm=',
    'npm': 'https://www.npmjs.com/package/',
    'NuGet': 'https://www.nuget.org/packages/',
    'Packagist': 'https://packagist.org/packages/',
    'Pub': 'https://pub.dev/packages/',
    'PyPI': 'https://pypi.org/project/',
    'Rocky Linux': 'https://pkgs.org/download/',
    'RubyGems': 'https://rubygems.org/gems/',
}

_OSV_TO_DEPS_ECOSYSTEMS_MAP = {
    'npm': 'npm',
    'Go': 'go',
    'Maven': 'maven',
    'PyPI': 'pypi',
    'NuGet': 'nuget',
    'crates.io': 'cargo'
}


def get(name: str) -> OrderedEcosystem | EnumerableEcosystem | None:
  """Get ecosystem helpers for a given ecosystem."""
  name, _, suffix = name.partition(':')
  ecosys = _ecosystems.get(name)
  if ecosys is None:
    return None
  return ecosys(suffix)


def normalize(ecosystem_name: str):
  return ecosystem_name.split(':')[0]


def remove_variants(ecosystem_name: str) -> str | None:
  result = None
  # For Ubuntu, remove ":Pro" and ":LTS"
  if ecosystem_name.startswith('Ubuntu'):
    result = ecosystem_name.replace(':Pro', '').replace(':LTS', '')

  return result


def add_matching_ecosystems(original_set: set[str]) -> set[str]:
  """
  For Linux distributions, some release versions may have different variants.
  For example, Ubuntu:22.04 is equivalent to Ubuntu:22.04:LTS.
  This function adds all matching ecosystems
  to the datastore to facilitate API queries.

  For example:
  - "Ubuntu:Pro:18.04:LTS" would also be "Ubuntu:18.04"

  Args:
    original_set: The original ecosystem set

  Returns:
    A new set with the added matching ecosystems.
  """
  new_set = set(original_set)
  for ecosystem in original_set:
    # For Ubuntu, remove ":Pro" and ":LTS"
    new_ecosystem = remove_variants(ecosystem)
    if new_ecosystem:
      new_set.add(new_ecosystem)
  return new_set


def is_supported_in_deps_dev(ecosystem_name: str) -> bool:
  return ecosystem_name in _OSV_TO_DEPS_ECOSYSTEMS_MAP


def map_ecosystem_to_deps_dev(ecosystem_name: str) -> str | None:
  return _OSV_TO_DEPS_ECOSYSTEMS_MAP.get(ecosystem_name)


def maybe_normalize_package_names(package_name: str, ecosystem: str) -> str:
  """Normalize package names as necessary."""
  if ecosystem == 'PyPI':
    # per https://peps.python.org/pep-0503/#normalized-names
    package_name = re.sub(r'[-_.]+', '-', package_name).lower()

  return package_name



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=line-too-long
# Many tests are ported from
# https://github.com/apache/maven/blob/c3cf29438e3d65d6ee5c5726f8611af99d9a649a/maven-artifact/src/test/java/org/apache/maven/artifact/versioning/ComparableVersionTest.java.
"""Maven ecosystem helper tests."""
import unittest
import vcr.unittest
import warnings

from . import maven
from .. import ecosystems


class MavenVersionTest(unittest.TestCase):
  """Maven version tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name

  def check_versions_order(self, *versions):
    """Check that our precedence logic matches the expected order."""
    parsed_versions = [maven.Version.from_string(v) for v in versions]

    # pylint: disable=consider-using-enumerate
    for i in range(len(parsed_versions)):
      for j in range(i + 1, len(parsed_versions)):
        self.assertLess(
            parsed_versions[i],
            parsed_versions[j],
            msg=f'Expected {versions[i]} < {versions[j]}')
        self.assertGreater(
            parsed_versions[j],
            parsed_versions[i],
            msg=f'Expected {versions[j]} > {versions[i]}')
        self.assertGreaterEqual(
            parsed_versions[j],
            parsed_versions[i],
            msg=f'Expected {versions[j]} >= {versions[i]}')

  def check_versions_equal(self, *versions):
    """Check that the provided versions are equivalent."""
    parsed_versions = [maven.Version.from_string(v) for v in versions]

    # pylint: disable=consider-using-enumerate
    for i in range(len(parsed_versions)):
      for j in range(i + 1, len(parsed_versions)):
        self.assertEqual(
            parsed_versions[i],
            parsed_versions[j],
            msg=f'Expected {versions[i]} == {versions[j]}')

  def test_normalize(self):
    """Test version normalization."""
    self.assertEqual('1', str(maven.Version.from_string('1.0.0')))
    self.assertEqual('1', str(maven.Version.from_string('1.ga')))
    self.assertEqual('1', str(maven.Version.from_string('1.final')))
    self.assertEqual('1', str(maven.Version.from_string('1.0')))
    self.assertEqual('1', str(maven.Version.from_string('1.')))
    self.assertEqual('1', str(maven.Version.from_string('1-')))
    self.assertEqual('1-foo', str(maven.Version.from_string('1.0.0-foo.0.0')))
    self.assertEqual('1', str(maven.Version.from_string('1.0.0-0.0.0')))
    self.assertEqual('1-1.foo-bar-1-baz-0.1',
                     str(maven.Version.from_string('1-1.foo-bar1baz-.1')))
    self.assertEqual('1-rc', str(maven.Version.from_string('1cr')))
    self.assertEqual('1-a', str(maven.Version.from_string('1a')))
    self.assertEqual('1-alpha-1', str(maven.Version.from_string('1a1')))
    self.assertEqual('1-beta-1', str(maven.Version.from_string('1b1')))
    self.assertEqual('1-c-1', str(maven.Version.from_string('1c1')))
    self.assertEqual('1-milestone-1', str(maven.Version.from_string('1m1')))
    self.assertEqual('1-1', str(maven.Version.from_string('1-ga-1')))

  def test_sort(self):
    """Basic sort tests."""
    # These tests are taken from the spec.
    unsorted = [
        '1', '1.1', '1-snapshot', '1', '1-sp', '1-foo2', '1-foo10', '1.foo',
        '1-foo', '1-1', '1.1', '1.ga', '1-ga', '1-0', '1.0', '1', '1-sp',
        '1-ga', '1-sp.1', '1-ga.1', '1-sp-1', '1-ga-1', '1-1', '1-a1',
        '1-alpha-1', '2', 'invalid', '0'
    ]

    sorted_versions = [
        str(v) for v in sorted(maven.Version.from_string(v) for v in unsorted)
    ]

    self.assertListEqual([
        'invalid', '0', '1-alpha-1', '1-alpha-1', '1-snapshot', '1', '1', '1',
        '1', '1', '1', '1', '1', '1.foo', '1-.1', '1-sp', '1-sp', '1-sp-1',
        '1-sp.1', '1-foo', '1-foo-2', '1-foo-10', '1-1', '1-1', '1-1', '1.1',
        '1.1', '2'
    ], sorted_versions)

  def test_versions_qualifiers(self):
    """Test qualifiers."""
    expected = [
        '1-alpha2snapshot', '1-alpha2', '1-alpha-123', '1-beta-2', '1-beta123',
        '1-m2', '1-m11', '1-rc', '1-cr2', '1-rc123', '1-SNAPSHOT', '1', '1-sp',
        '1-sp2', '1-sp123', '1-abc', '1-def', '1-pom-1', '1-1-snapshot', '1-1',
        '1-2', '1-123'
    ]
    self.check_versions_order(*expected)

  def test_versions_number(self):
    """Test numbers."""
    # Taken from Maven's tests.
    expected = [
        '2.0', '2-1', '2.0.a', '2.0.0.a', '2.0.2', '2.0.123', '2.1.0', '2.1-a',
        '2.1b', '2.1-c', '2.1-1', '2.1.0.1', '2.2', '2.123', '11.a2', '11.a11',
        '11.b2', '11.b11', '11.m2', '11.m11', '11', '11.a', '11b', '11c', '11m'
    ]
    self.check_versions_order(*expected)

  def test_versions_order(self):
    """More ordering tests."""
    self.check_versions_order('1', '2')
    self.check_versions_order('1.5', '2')
    self.check_versions_order('1', '2.5')
    self.check_versions_order('1.0', '1.1')
    self.check_versions_order('1.1', '1.2')
    self.check_versions_order('1.0.0', '1.1')
    self.check_versions_order('1.0.1', '1.1')
    self.check_versions_order('1.1', '1.2.0')

    self.check_versions_order('1.0-alpha-1', '1.0')
    self.check_versions_order('1.0-alpha-1', '1.0-alpha-2')
    self.check_versions_order('1.0-alpha-1', '1.0-beta-1')

    self.check_versions_order('1.0-beta-1', '1.0-SNAPSHOT')
    self.check_versions_order('1.0-SNAPSHOT', '1.0')
    self.check_versions_order('1.0-alpha-1-SNAPSHOT', '1.0-alpha-1')

    self.check_versions_order('1.0', '1.0-1')
    self.check_versions_order('1.0-1', '1.0-2')
    self.check_versions_order('1.0.0', '1.0-1')

    self.check_versions_order('2.0-1', '2.0.1')
    self.check_versions_order('2.0.1-klm', '2.0.1-lmn')
    self.check_versions_order('2.0.1', '2.0.1-xyz')

    self.check_versions_order('2.0.1', '2.0.1-123')
    self.check_versions_order('2.0.1-xyz', '2.0.1-123')

  def test_versions_order_mng_5568(self):
    """Regression test for MNG 5568."""
    a = '6.1.0'
    b = '6.1.0rc3'
    c = '6.1H.5-beta'

    self.check_versions_order(b, a)
    self.check_versions_order(b, c)
    self.check_versions_order(a, c)

  def test_versions_order_mng_6572(self):
    """Regression test for MNG 6572."""
    a = '20190126.230843'
    b = '1234567890.12345'
    c = '123456789012345.1H.5-beta'
    d = '12345678901234567890.1H.5-beta'

    self.check_versions_order(a, b)
    self.check_versions_order(b, c)
    self.check_versions_order(a, c)
    self.check_versions_order(c, d)
    self.check_versions_order(b, d)
    self.check_versions_order(a, d)

  def test_versions_equal(self):
    """Test versions that should be considered equal."""
    self.check_versions_equal('1', '1')
    self.check_versions_equal('1', '1.0')
    self.check_versions_equal('1', '1.0.0')
    self.check_versions_equal('1.0', '1.0.0')
    self.check_versions_equal('1', '1-0')
    self.check_versions_equal('1', '1.0-0')
    self.check_versions_equal('1.0', '1.0-0')
    # no separator between number and character
    self.check_versions_equal('1a', '1-a')
    self.check_versions_equal('1a', '1.0-a')
    self.check_versions_equal('1a', '1.0.0-a')
    self.check_versions_equal('1.0a', '1-a')
    self.check_versions_equal('1.0.0a', '1-a')
    self.check_versions_equal('1x', '1-x')
    self.check_versions_equal('1x', '1.0-x')
    self.check_versions_equal('1x', '1.0.0-x')
    self.check_versions_equal('1.0x', '1-x')
    self.check_versions_equal('1.0.0x', '1-x')

    # aliases
    self.check_versions_equal('1ga', '1')
    self.check_versions_equal('1release', '1')
    self.check_versions_equal('1final', '1')
    self.check_versions_equal('1cr', '1rc')

    # special 'aliases' a, b and m for alpha, beta and milestone
    self.check_versions_equal('1a1', '1-alpha-1')
    self.check_versions_equal('1b2', '1-beta-2')
    self.check_versions_equal('1m3', '1-milestone-3')

    # case insensitive
    self.check_versions_equal('1X', '1x')
    self.check_versions_equal('1A', '1a')
    self.check_versions_equal('1B', '1b')
    self.check_versions_equal('1M', '1m')
    self.check_versions_equal('1Ga', '1')
    self.check_versions_equal('1GA', '1')
    self.check_versions_equal('1RELEASE', '1')
    self.check_versions_equal('1release', '1')
    self.check_versions_equal('1RELeaSE', '1')
    self.check_versions_equal('1Final', '1')
    self.check_versions_equal('1FinaL', '1')
    self.check_versions_equal('1FINAL', '1')
    self.check_versions_equal('1Cr', '1Rc')
    self.check_versions_equal('1cR', '1rC')
    self.check_versions_equal('1m3', '1Milestone3')
    self.check_versions_equal('1m3', '1MileStone3')
    self.check_versions_equal('1m3', '1MILESTONE3')

    self.check_versions_equal('1', '01', '001')
    self.check_versions_equal('1.1', '1.01', '1.001')
    self.check_versions_equal('1-1', '1-01', '1-001')

  def test_version_zero(self):
    """Test comparison and equality with versions 0.0.0"""
    self.check_versions_equal('0.0.0', '0.0', '0')
    self.check_versions_equal('0.0.0-0.0.0', '0-final-ga', '0')
    self.check_versions_order('0', '1')

    # actual case from com.graphql-java:graphql-java
    self.check_versions_order('0.0.0-2021-05-17T01-01-51-5ec03a8b', '20.0.0')

  def test_version_ge_le(self):
    """Test >= and <=."""
    self.assertGreaterEqual(
        maven.Version.from_string('1.10.0'), maven.Version.from_string('1.2.0'))
    self.assertLessEqual(
        maven.Version.from_string('1.2.0'), maven.Version.from_string('1.10.0'))


class MavenEcosystemTest(vcr.unittest.VCRTestCase):
  """Maven ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('Maven')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('1.36.0',
                       ecosystem.next_version('io.grpc:grpc-core', '1.35.1'))
      self.assertEqual('0.7.0', ecosystem.next_version('io.grpc:grpc-core',
                                                       '0'))
      with self.assertRaises(ecosystems.EnumerateError):
        ecosystem.next_version('blah:doesnotexist123456', '1')

  def test_enumerate(self):
    """Test enumerate."""
    ecosystem = ecosystems.get('Maven')
    self.assertEqual(['10.0', '10.0.1', '11.0-rc1', '11.0'],
                     ecosystem.enumerate_versions(
                         'com.google.guava:guava', '10.0',
                         last_affected='11.0'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bug helper tests."""

import unittest

from . import bug


class NormalizeTest(unittest.TestCase):
  """Normalize tests."""

  def test_normalize(self):
    """Test version normalization."""
    tags = [
        '1.0',
        '22.3rc1',
        '1.2.3.4.5-rc1',
        '.1',
        '0.1.11.1',
        'project-123-1',
        'project-123-1RC',
        'project-123-1RC5',
        'arc-20200101',
        'php-8.0.0beta',
        'php-8.0.0beta4',
        'v6.0.0-alpha1',
        'android-10.0.0_r10',
    ]

    self.assertListEqual([
        '1-0',
        '22-3-rc1',
        '1-2-3-4-5-rc1',
        '1',
        '0-1-11-1',
        '123-1',
        '123-1-RC',
        '123-1-RC5',
        '20200101',
        '8-0-0-beta',
        '8-0-0-beta4',
        '6-0-0-alpha1',
        '10-0-0-10',
    ], bug.normalize_tags(tags))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystems base classes."""
from abc import ABC, abstractmethod
from typing import Any
from warnings import deprecated
import bisect
import requests
from urllib.parse import quote

from . import config


class OrderedEcosystem(ABC):
  """Ecosystem helper that supports comparison between versions."""

  def __init__(self, suffix: str | None = None):
    """init method for all ecosystem helpers.
    
    `suffix` is optionally used on ecosystems that use them.
    e.g. Alpine:v3.16 would use suffix='v3.16'
    """
    self.suffix = suffix

  @abstractmethod
  def sort_key(self, version: str) -> Any:
    """Comparable key for a version.
    
    If the version string is invalid, return a very large version.
    """

  def sort_versions(self, versions: list[str]):
    """Sort versions."""
    versions.sort(key=self.sort_key)


class EnumerateError(Exception):
  """Non-retryable version enumeration error."""


class EnumerableEcosystem(OrderedEcosystem, ABC):
  """Ecosystem helper that supports version enumeration."""

  @abstractmethod
  def enumerate_versions(self,
                         package: str,
                         introduced: str | None,
                         fixed: str | None = None,
                         last_affected: str | None = None,
                         limits: list[str] | None = None) -> list[str]:
    """Enumerate known versions of a package in a given version range."""

  def _before_limits(self, version: str, limits: list[str] | None) -> bool:
    """Return whether the given version is before any limits."""
    if not limits or '*' in limits:
      return True

    return any(
        self.sort_key(version) < self.sort_key(limit) for limit in limits)

  def _get_affected_versions(self, versions: list[str], introduced: str | None,
                             fixed: str | None, last_affected: str | None,
                             limits: list[str] | None) -> list[str]:
    """Get affected versions.

    Args:
      versions: a list of version strings.
      introduced: a version string.
      fixed: a version string.
      last_affected: a version string.
      limits: a version string.

    Returns:
      A list of affected version strings.
    """
    parsed_versions = [self.sort_key(v) for v in versions]

    if introduced == '0':
      introduced = None

    if introduced:
      introduced = self.sort_key(introduced)
      start_idx = bisect.bisect_left(parsed_versions, introduced)
    else:
      start_idx = 0

    if fixed:
      fixed = self.sort_key(fixed)
      end_idx = bisect.bisect_left(parsed_versions, fixed)
    elif last_affected:
      last_affected = self.sort_key(last_affected)
      end_idx = bisect.bisect_right(parsed_versions, last_affected)
    else:
      end_idx = len(versions)

    affected = versions[start_idx:end_idx]
    return [v for v in affected if self._before_limits(v, limits)]

  @deprecated('Avoid using this method. '
              'It is provided only to maintain existing tooling.')
  def next_version(self, package: str, version: str) -> str | None:
    """Get the next version after the given version."""
    versions = self.enumerate_versions(package, version, fixed=None)
    # Check if the key used for sorting is equal as sometimes different
    # strings could evaluate to the same version.
    if versions and self.sort_key(versions[0]) != self.sort_key(version):
      # Version does not exist, so use the first one that would sort
      # after it (which is what enumerate_versions returns).
      return versions[0]

    if len(versions) > 1:
      return versions[1]

    return None


class DepsDevMixin(EnumerableEcosystem, ABC):
  """deps.dev mixin."""

  _DEPS_DEV_PACKAGE_URL = \
      'https://api.deps.dev/v3alpha/systems/{system}/packages/{package}'

  @property
  @abstractmethod
  def deps_dev_system(self) -> str:
    """The deps.dev system name."""

  def _deps_dev_enumerate(self,
                          package,
                          introduced,
                          fixed=None,
                          last_affected=None,
                          limits=None):
    """Use deps.dev to get list of versions."""
    url = self._DEPS_DEV_PACKAGE_URL.format(
        system=self.deps_dev_system, package=quote(package, safe=''))
    response = requests.get(url, timeout=config.timeout)
    if response.status_code == 404:
      raise EnumerateError(f'Package {package} not found')
    if response.status_code != 200:
      raise RuntimeError(
          f'Failed to get {self.deps_dev_system} versions for {package} with: '
          f'{response.status_code}')
    response = response.json()
    versions = [v['versionKey']['version'] for v in response['versions']]
    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Request helpers tests"""

import unittest
import types
from unittest import mock

from .request_helper import RequestHelper
from .cache import InMemoryCache


class RequestHelperTests(unittest.TestCase):
  """get_next_version tests."""

  @mock.patch('requests.sessions.Session.get')
  def test_cache(self, session_get: mock.MagicMock):
    """Test cache works as expected."""

    test_response = types.SimpleNamespace()
    test_response.status_code = 200
    test_response.text = 'test123'
    test_url = 'https://example.com'

    session_get.return_value = test_response

    cache = InMemoryCache()
    request_helper = RequestHelper(cache)

    example_result = request_helper.get(test_url)
    self.assertEqual(example_result, test_response.text)
    session_get.assert_called_once_with(test_url)

    cached_response = request_helper.get(test_url)
    self.assertEqual(example_result, cached_response)
    session_get.assert_called_once()  # Still only called once


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bioconductor helpers."""

import requests

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .semver_ecosystem_helper import SemverLike


class Bioconductor(EnumerableEcosystem, SemverLike):
  """Bioconductor ecosystem helpers."""

  # Use the Posit Public Package Manager API to pull both the current and
  # older versions for a specific package since Bioconductor doesn't natively
  # support this functionality.
  _API_PACKAGE_URL_POSIT_BIOCONDUCTOR = \
    'https://packagemanager.posit.co/__api__/repos/4/packages/' + \
    '{package}?bioc_version={bioc_version}'
  _API_BIOC_VERSIONS_URL = 'https://packagemanager.posit.co/__api__/status'

  def get_bioc_versions(self):
    """
    get latest Bioconductor versions
    """
    response = requests.get(self._API_BIOC_VERSIONS_URL, timeout=60)
    data = response.json()
    if response.status_code == 404:
      raise RuntimeError('Failed to get Bioconductor versions')

    return [bioc['bioc_version'] for bioc in data['bioc_versions']]

  def _enumerate_versions(self,
                          url,
                          bioc_versions,
                          package,
                          introduced,
                          fixed=None,
                          last_affected=None,
                          limits=None):
    """Helper method to enumerate versions from a specific URL."""

    versions = []
    for version in bioc_versions:
      response = requests.get(
          url.format(package=package, bioc_version=version),
          timeout=config.timeout)
      if response.status_code == 404:
        continue

      if response.status_code != 200:
        raise RuntimeError(
            f'Failed to get R versions for {package} with: {response.text}')

      response = response.json()
      if 'version' in response:
        versions += [response['version']]

    if len(versions) == 0:
      raise EnumerateError(f'Package {package} not found')

    self.sort_versions(versions)
    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    # Refresh to ensure any new Bioconductor versions are included
    bioc_versions = self.get_bioc_versions()
    versions = self._enumerate_versions(
        self._API_PACKAGE_URL_POSIT_BIOCONDUCTOR, bioc_versions, package,
        introduced, fixed, last_affected, limits)

    if versions is None:
      raise EnumerateError(f'Package {package} not found')

    return versions



"""Module for parsing converting CSAF to OSV data"""
import re
from dataclasses import field, dataclass, InitVar
import json
from typing import Literal
import requests
from jsonschema import validate
from redhat_osv.csaf import Remediation, CSAF

# Update this if verified against a later version
SCHEMA_VERSION = "1.7.0"
# This assumes the datetime being formatted is in UTC
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
# Go package advisory reference prefix
PKG_GO_DEV_VULN = "https://pkg.go.dev/vuln/"
REDHAT_ADVISORY_URL = "https://access.redhat.com/errata/"
# Other common advisory prefixes in Red Hat Advisories
ADVISORY_URL_PREFIXES = (
    PKG_GO_DEV_VULN,
    "https://www.cve.org/CVERecord",
    "https://nvd.nist.gov/vuln/detail/",
    "https://www.kb.cert.org/vuls/id/",
    "https://github.com/advisories/",
)


class OSVEncoder(json.JSONEncoder):
    """Encodes OSV objects into JSON format"""

    def default(self, o):
        if isinstance(o, Event):
            return o.encode_json()
        return o.__dict__


@dataclass
class Event:
    """
    Class to hold event information for a Range. Advisories for Red Hat RPM based products always
    assume all previous versions are affected.
    """

    event_type: Literal["introduced", "fixed"]
    version: str = "0"
    introduced: Literal["introduced"] = "introduced"
    fixed: Literal["fixed"] = "fixed"

    def __post_init__(self):
        if self.event_type not in (self.introduced, self.fixed):
            raise ValueError(
                f"Expected one of {(self.introduced, self.fixed)} for type. "
                f"Got {self.event_type}")

    def encode_json(self):
        """
        Custom JSON encoding for event type which changes attribute name depending on the type of
        event eg. introduced or fixed
        """
        if self.event_type == "introduced":
            return {"introduced": self.version}
        if self.event_type == "fixed":
            return {"fixed": self.version}
        raise ValueError("Unexpected event_type for Event")


@dataclass
class Range:
    """
    Class to hold range information for a Package. Ecosystem here refers to RPM versions as defined
    in https://github.com/rpm-software-management/rpm/blob/master/rpmio/rpmvercmp.c
    """

    fixed: InitVar[str]
    type: str = field(init=False)
    events: list[Event] = field(init=False)

    def __post_init__(self, fixed):
        self.events = [Event("introduced"), Event("fixed", fixed)]
        self.type = "ECOSYSTEM"


@dataclass
class Package:
    """
    Class to hold package data for an Affect.
    Expects an ecosystem string that starts with CPE_PATTERN.
    Replaces the CPE prefix 'redhat' part with 'Red Hat'
    to match more closely with other ecosystem identifiers in the OSV database.
    Also removes version and qualifiers from the CSAF remediation PURL
    """

    cpe_pattern: re.Pattern = field(init=False,
                                    default=re.compile(r"cpe:/[oa]:(redhat)"))
    name: str
    ecosystem: str
    purl: str

    def __post_init__(self):
        if not self.cpe_pattern.match(self.ecosystem):
            raise ValueError(f"Got unsupported ecosystem: {self.ecosystem}")
        self.ecosystem = f"Red Hat{self.cpe_pattern.split(self.ecosystem, maxsplit=1)[-1]}"
        if "@" in self.purl:
            version_index = self.purl.index("@")
            self.purl = self.purl[:version_index]



@dataclass
class Affected:
    """
    Class to hold affected data for a Vulnerability
    """

    package: Package
    ranges: list[Range]


# pylint: disable=too-many-instance-attributes
# This class is directly encoded into OSV json so has one instance attribute for each OSV property
class OSV:
    """
    Class to convert CSAF data to OSV
    """

    def __init__(self, csaf_data: CSAF, modified: str, published: str = ""):
        self.schema_version = SCHEMA_VERSION

        self.id = ""

        # This attribute is declared after id to make the resulting JSON human-readable. It can only
        # be populated after reading the csaf vulnerabilities and references sections.
        self.related: list[str] = []
        self.upstream: list[str] = []

        if published:
            self.published = published
        else:
            self.published = modified
        self.modified = modified

        self.summary = csaf_data.title

        # Set severity to the CVSS of the highest CVSSv3 base score
        vulnerability_scores: dict[str, str] = {}
        for vulnerability in csaf_data.vulnerabilities:
            if not vulnerability.cvss_v3_vector or not vulnerability.cvss_v3_base_score:
                continue
            vulnerability_scores[
                vulnerability.cvss_v3_base_score] = vulnerability.cvss_v3_vector
        if vulnerability_scores:
            highest_score = sorted(vulnerability_scores.keys())[-1]
            self.severity = [{
                "type": "CVSS_V3",
                "score": vulnerability_scores[highest_score]
            }]

        self.affected: list[Affected] = []

        # Deduplicate arch specific remediations
        unique_packages: dict[str: tuple[str: str]] = {}

        for vulnerability in csaf_data.vulnerabilities:
            self.upstream.append(vulnerability.cve_id)
            for remediation in vulnerability.remediations:
                # Safety check for when we start processing non-rpm content
                if not remediation.purl.startswith("pkg:rpm/"):
                    package = Package(remediation.component, remediation.cpe, remediation.purl)
                    ranges = [Range(remediation.fixed_version)]
                    self.affected.append(Affected(package, ranges))
                else:
                    # Each RPM version in RHEL has a trailing '.<arch>', remove those to avoid
                    # problems comparing the same package from different archs
                    version_arch_split = remediation.fixed_version.rsplit(".", 1)
                    # CPE's are URI percent encoded and '&' is a reserved character so it should
                    # never appear in a CPE without being percent encoded.
                    unique_packages[remediation.cpe + "&" + remediation.component] = (
                        version_arch_split[0], remediation.purl,
                    )

        # Add all the RPM packages without arch suffixes
        for package_key, version_purl in unique_packages.items():
            package_key_parts = package_key.split("&", 1)
            cpe = package_key_parts[0]
            component = package_key_parts[1]
            package = Package(component, cpe, version_purl[1])
            ranges = [Range(version_purl[0])]
            self.affected.append(Affected(package, ranges))

        self.references = self._convert_references(csaf_data)

    def _convert_references(self, csaf) -> list[dict[str, str]]:
        """
        CSAF has references for an advisory and each vulnerability has references as well.
        Collect this into a single references list for OSV and deduplicate them.
        """
        references: dict[str, str] = {}
        for reference in csaf.references:
            # This will capture both the Advisory URL and the CSAF document for the advisory
            if reference["category"] == "self":
                if reference["summary"].startswith(REDHAT_ADVISORY_URL):
                    self.id = reference["summary"].removeprefix(
                        REDHAT_ADVISORY_URL)
                references[reference["url"]] = "ADVISORY"
            else:
                references[reference[
                    "url"]] = self._get_reference_type_and_add_go_related(
                        reference)
        for vulnerability in csaf.vulnerabilities:
            for reference in vulnerability.references:
                # This captures the CVE specific information
                if reference["category"] == "self":
                    references[reference["url"]] = "REPORT"
                else:
                    references[reference[
                        "url"]] = self._get_reference_type_and_add_go_related(
                            reference)
        return [{"type": t, "url": u} for u, t in references.items()]

    def _get_reference_type_and_add_go_related(
            self, reference: dict[str, str]) -> str:
        """
        Convert references from CSAF into typed referenced in OSV
        Also make sure to add a related entry for any GO advisory references found
        """
        reference_url = reference["url"]
        if reference_url.startswith(ADVISORY_URL_PREFIXES):
            if reference_url.startswith(PKG_GO_DEV_VULN):
                self.related.append(reference_url.removeprefix(PKG_GO_DEV_VULN))
            return "ADVISORY"
        if reference_url.startswith("https://bugzilla.redhat.com/show_bug.cgi"):
            return "REPORT"
        return "ARTICLE"


class RedHatConverter:
    """
    Class which converts and validates a CSAF string to an OSV string
    """
    SCHEMA = (
        f"https://raw.githubusercontent.com/ossf/osv-schema/v{SCHEMA_VERSION}"
        "/validation/schema.json")
    REQUEST_TIMEOUT = 60

    def __init__(self):
        schema_content = requests.get(self.SCHEMA, timeout=self.REQUEST_TIMEOUT)
        self.osv_schema = schema_content.json()

    def convert(self,
                csaf_content: str,
                modified: str,
                published: str = "") -> tuple[str, str]:
        """
        Converts csaf_content json string into an OSV json string
        returns an OSV ID and the json string content of the OSV file
        the json string content will be empty if no content is applicable
        throws a validation error in the schema doesn't validate correctly.
        The modified value for osv is passed in so it matches what's in all.json
        Raises ValueError is CSAF file can't be parsed
        """
        csaf = CSAF(csaf_content)
        osv = OSV(csaf, modified, published)

        # We convert from an OSV object to a JSON string here in order to use the OSVEncoder
        # Once we OSV json string data we validate it using the OSV schema
        osv_content = json.dumps(osv, cls=OSVEncoder, indent=2)
        osv_data = json.loads(osv_content)
        validate(osv_data, schema=self.osv_schema)

        return osv.id, osv_content



# Copyright 2025 Google LLC
# Copyright 2023 Fraser Tweedale
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Haskell ecosystem helper tests."""

import warnings

import vcr.unittest

from .. import ecosystems


class HackageEcosystemTest(vcr.unittest.VCRTestCase):
  """Hackage ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('Hackage')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('1.0.0.0', ecosystem.next_version('aeson', '0.11.3.0'))
      self.assertEqual('1.0.1.0', ecosystem.next_version('aeson', '1.0.0.0'))
      self.assertEqual('0.1.26.0', ecosystem.next_version('jose', '0'))
      with self.assertRaises(ecosystems.EnumerateError):
        ecosystem.next_version('doesnotexist123456', '1')

  def test_sort_key(self):
    """Test sort_key."""
    ecosystem = ecosystems.get('Hackage')
    self.assertGreater(
        ecosystem.sort_key('1-20-0'), ecosystem.sort_key('1.20.0'))

    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('1-20-0'), ecosystem.sort_key('1.20.0'))
    self.assertLessEqual(
        ecosystem.sort_key('1.20.0'), ecosystem.sort_key('1-20-0'))


class GHCEcosystemTest(vcr.unittest.VCRTestCase):
  """GHC ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('GHC')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('0.29', ecosystem.next_version('GHC', '0'))
      self.assertEqual('7.0.4', ecosystem.next_version('GHC', '7.0.4-rc1'))
      # 7.0.4 is the last of the hardcoded versions
      self.assertEqual('7.2.1', ecosystem.next_version('GHC', '7.0.4'))

      # The whole GHC ecosystem is versioned together.  Enumeration ignores
      # package/component name.  Therefore this should NOT raise:
      ecosystem.next_version('doesnotexist123456', '1')



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pub ecosystem helper."""

import functools
import json

from . import config
from .ecosystems_base import EnumerableEcosystem, EnumerateError
from .. import semver_index
from ..request_helper import RequestError, RequestHelper

# Pub follows SemVer 2.0.0-rc.1:
# https://semver.org/spec/v2.0.0-rc.1.html
# Differences from SemVer 2.0.0 are described at
# https://pub.dev/packages/pub_semver
# Only difference that affects us is the inclusion of build suffix when
# ordering.
#  - Build suffixes are parsed the same as pre-release suffixes.
#  - Pre-release suffixes are evaluated before build suffixes.
#  - Versions with build suffixes come after versions without.
#    e.g. 1.0.0-pre < 1.0.0-pre+build < 1.0.0 < 1.0.0+build
# SemVer 2.0.0-rc.1 also does not explicitly disallow empty identifiers or
# leading 0s on numeric identifiers, but our SemVer implementation also will
# parse these cases.


@functools.total_ordering
class Version:
  """Pub version."""

  def __init__(self, semver):
    # Tuple comparison of (semver,) or (semver, build) is consistent with
    # desired ordering.
    if semver.build:
      # Reuse normalize_prerelease to so that the
      # build suffix can be compared lexicographically.
      self._version = (semver, semver_index.normalize_prerelease(semver.build))
    else:
      self._version = (semver,)

  def __lt__(self, other):
    return self._version < other._version

  def __eq__(self, other):
    return self._version == other._version

  @classmethod
  def from_string(cls, str_version):
    # If version is not valid, it is most likely an invalid input
    # version then sort it to the last/largest element
    try:
      return Version(semver_index.parse(str_version))
    except ValueError:
      return Version(semver_index.parse('999999'))


class Pub(EnumerableEcosystem):
  """Pub ecosystem"""

  _API_PACKAGE_URL = 'https://pub.dev/api/packages/{package}'

  def sort_key(self, version):
    """Sort key."""
    return Version.from_string(version)

  def enumerate_versions(self,
                         package,
                         introduced,
                         fixed=None,
                         last_affected=None,
                         limits=None):
    """Enumerate versions."""
    url = self._API_PACKAGE_URL.format(package=package.lower())
    request_helper = RequestHelper(config.shared_cache)
    try:
      text_response = request_helper.get(url)
    except RequestError as ex:
      if ex.response.status_code == 404:
        raise EnumerateError(f'Package {package} not found') from ex
      raise RuntimeError('Failed to get Pub versions for '
                         f'{package} with: {ex.response.text}') from ex

    response = json.loads(text_response)
    versions: list[str] = [x['version'] for x in response['versions']]
    self.sort_versions(versions)

    return self._get_affected_versions(versions, introduced, fixed,
                                       last_affected, limits)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=line-too-long
# Many tests are ported from
# https://github.com/NuGet/NuGet.Client/blob/dev/test/NuGet.Core.Tests/NuGet.Versioning.Test/VersionComparerTests.cs
"""NuGet ecosystem helper tests."""

import unittest
import vcr.unittest
import warnings

from . import nuget
from .. import ecosystems


class NuGetVersionTest(unittest.TestCase):
  """NuGet version tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name

  def check_order(self, comparison, first, second):
    """Check order."""
    comparison(
        nuget.Version.from_string(first), nuget.Version.from_string(second))

  def test_equals(self):
    """Test version equals."""
    self.check_order(self.assertEqual, '1.0.0', '1.0.0')
    self.check_order(self.assertEqual, '1.0.0-BETA', '1.0.0-beta')
    self.check_order(self.assertEqual, '1.0.0-BETA+AA', '1.0.0-beta+aa')
    self.check_order(self.assertEqual, '1.0.0-BETA.X.y.5.77.0+AA',
                     '1.0.0-beta.x.y.5.77.0+aa')
    self.check_order(self.assertEqual, '1.0.0', '1.0.0+beta')

    self.check_order(self.assertEqual, '1.0', '1.0.0.0')
    self.check_order(self.assertEqual, '1.0+test', '1.0.0.0')
    self.check_order(self.assertEqual, '1.0.0.1-1.2.A', '1.0.0.1-1.2.a+A')
    self.check_order(self.assertEqual, '1.0.01', '1.0.1.0')

  def test_not_equals(self):
    """Test version not equals."""
    self.check_order(self.assertNotEqual, '1.0', '1.0.0.1')
    self.check_order(self.assertNotEqual, '1.0+test', '1.0.0.1')
    self.check_order(self.assertNotEqual, '1.0.0.1-1.2.A', '1.0.0.1-1.2.a.A+A')
    self.check_order(self.assertNotEqual, '1.0.01', '1.0.1.2')
    self.check_order(self.assertNotEqual, '0.0.0', '1.0.0')
    self.check_order(self.assertNotEqual, '1.1.0', '1.0.0')
    self.check_order(self.assertNotEqual, '1.0.1', '1.0.0')
    self.check_order(self.assertNotEqual, '1.0.0-BETA', '1.0.0-beta2')
    self.check_order(self.assertNotEqual, '1.0.0+AA', '1.0.0-beta+aa')
    self.check_order(self.assertNotEqual, '1.0.0-BETA.X.y.5.77.0+AA',
                     '1.0.0-beta.x.y.5.79.0+aa')

  def test_less(self):
    """Test version less."""
    self.check_order(self.assertLess, '0.0.0', '1.0.0')
    self.check_order(self.assertLess, '1.0.0', '1.1.0')
    self.check_order(self.assertLess, '1.0.0', '1.0.1')
    self.check_order(self.assertLess, '1.999.9999', '2.1.1')
    self.check_order(self.assertLess, '1.0.0-BETA', '1.0.0-beta2')
    self.check_order(self.assertLess, '1.0.0-beta+AA', '1.0.0+aa')
    self.check_order(self.assertLess, '1.0.0-BETA', '1.0.0-beta.1+AA')
    self.check_order(self.assertLess, '1.0.0-BETA.X.y.5.77.0+AA',
                     '1.0.0-beta.x.y.5.79.0+aa')
    self.check_order(self.assertLess, '1.0.0-BETA.X.y.5.79.0+AA',
                     '1.0.0-beta.x.y.5.790.0+abc')

    self.check_order(self.assertLess, '1.0.0', '1.0.0.1')
    self.check_order(self.assertLess, '1.0.0.1-alpha', '1.0.0.1-pre')
    self.check_order(self.assertLess, '1.0.0-pre', '1.0.0.1-alpha')
    self.check_order(self.assertLess, '1.0.0', '1.0.0.1-alpha')
    self.check_order(self.assertLess, '0.9.9.1', '1.0.0')

  def test_ge_le(self):
    """Test version >=/<=."""
    self.check_order(self.assertGreaterEqual, '1.10.0', '1.2.0')
    self.check_order(self.assertLessEqual, '1.2.0', '1.10.0')


class NuGetEcosystemTest(vcr.unittest.VCRTestCase):
  """NuGet ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('NuGet')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('3.0.1',
                       ecosystem.next_version('NuGet.Server.Core', '3.0.0'))
      self.assertEqual('3.0.0.4001',
                       ecosystem.next_version('Castle.Core', '3.0.0.3001'))
      self.assertEqual('3.1.0-RC',
                       ecosystem.next_version('Castle.Core', '3.0.0.4001'))
      self.assertEqual('2.1.0-dev-00668',
                       ecosystem.next_version('Serilog', '2.1.0-dev-00666'))
      with self.assertRaises(ecosystems.EnumerateError):
        ecosystem.next_version('doesnotexist123456', '1')

  def test_sort_key(self):
    ecosystem = ecosystems.get('NuGet')
    # Tests invalid input versions
    self.assertGreater(
        ecosystem.sort_key('1.4.0rc3'), ecosystem.sort_key('3.0.0.4001'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RubyGems ecosystem helper tests."""

import vcr.unittest
import warnings

from .. import ecosystems


class RubyGemsEcosystemTest(vcr.unittest.VCRTestCase):
  """RubyGems ecosystem helper tests."""

  def test_next_version(self):
    """Test next_version."""
    ecosystem = ecosystems.get('RubyGems')
    with warnings.catch_warnings():
      # Filter the DeprecationWarning from next_version
      warnings.filterwarnings('ignore', 'Avoid using this method')
      self.assertEqual('0.8.0', ecosystem.next_version('rails', '0'))
      self.assertEqual('0.9.5', ecosystem.next_version('rails', '0.9.4.1'))
      self.assertEqual('2.3.8.pre1', ecosystem.next_version('rails', '2.3.7'))
      self.assertEqual('4.0.0.rc1',
                       ecosystem.next_version('rails', '4.0.0.beta1'))
      self.assertEqual('5.0.0.racecar1',
                       ecosystem.next_version('rails', '5.0.0.beta4'))
      with self.assertRaises(ecosystems.EnumerateError):
        ecosystem.next_version('doesnotexist123456', '1')

  def test_sort_key(self):
    """Test sort_key with invalid versions"""
    ecosystem = ecosystems.get('RubyGems')
    self.assertGreater(
        ecosystem.sort_key('invalid'), ecosystem.sort_key('4.0.0.rc1'))
    self.assertGreater(
        ecosystem.sort_key('v3.1.1'), ecosystem.sort_key('4.0.0.rc1'))
    # Check >= / <= methods
    self.assertGreaterEqual(
        ecosystem.sort_key('1.10.0.rc1'), ecosystem.sort_key('1.2.0.rc1'))
    self.assertLessEqual(
        ecosystem.sort_key('1.2.0.rc1'), ecosystem.sort_key('1.10.0.rc1'))



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GCP Logging Helpers."""

import logging
import os

from google.cloud import logging as google_logging


class _ErrorReportingFilter:
  """
  A logging filter that adds necessary JSON fields to error logs so that they
  can be picked up by Error Reporting.
  
  https://cloud.google.com/error-reporting/docs/formatting-error-messages#log-text
  https://docs.python.org/3/howto/logging-cookbook.html#using-filters-to-impart-contextual-information
  """

  def __init__(self, service_name: str) -> None:
    self.service_name = service_name

  def filter(self, record: logging.LogRecord) -> bool:
    """Add the error reporting fields to json_fields."""
    if not hasattr(record, 'json_fields'):
      record.json_fields = {}

    if record.levelno >= logging.ERROR and not record.exc_info:
      record.json_fields.update({
          '@type':
              'type.googleapis.com/google.devtools.clouderrorreporting.v1beta1.ReportedErrorEvent',  # pylint: disable=line-too-long
          'serviceContext': {
              'service': self.service_name,
          },
          'context': {
              'reportLocation': {
                  'filePath': record.pathname,
                  'lineNumber': record.lineno,
                  'functionName': record.funcName,
              }
          },
      })

    return True


def setup_gcp_logging(service_name):
  """Set up GCP logging and error reporting."""

  if (os.getenv('KUBERNETES_SERVICE_HOST') is not None or
      os.getenv('K_SERVICE') is not None):
    logging_client = google_logging.Client()
    logging_client.setup_logging()

  logging.getLogger().addFilter(_ErrorReportingFilter(service_name))
  logging.getLogger().setLevel(logging.INFO)

  # Suppress noisy logs in some of our dependencies.
  logging.getLogger('google.api_core.bidi').setLevel(logging.ERROR)
  logging.getLogger('google.cloud.pubsub_v1.subscriber._protocol.'
                    'streaming_pull_manager').setLevel(logging.ERROR)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vulnerability.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13vulnerability.proto\x12\x03osv\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"y\n\x06\x43ommit\x12\'\n\trepo_type\x18\x01 \x01(\x0e\x32\x14.osv.Commit.RepoType\x12\x10\n\x08repo_url\x18\x02 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x03 \x01(\t\"$\n\x08RepoType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03GIT\x10\x01\"8\n\x07Package\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tecosystem\x18\x02 \x01(\t\x12\x0c\n\x04purl\x18\x03 \x01(\t\"P\n\x05\x45vent\x12\x12\n\nintroduced\x18\x01 \x01(\t\x12\r\n\x05\x66ixed\x18\x02 \x01(\t\x12\r\n\x05limit\x18\x03 \x01(\t\x12\x15\n\rlast_affected\x18\x04 \x01(\t\"\x8d\x01\n\x05Range\x12\x1d\n\x04type\x18\x01 \x01(\x0e\x32\x0f.osv.Range.Type\x12\x0c\n\x04repo\x18\x02 \x01(\t\x12\x1a\n\x06\x65vents\x18\x03 \x03(\x0b\x32\n.osv.Event\";\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03GIT\x10\x01\x12\n\n\x06SEMVER\x10\x02\x12\r\n\tECOSYSTEM\x10\x03\"\xe1\x01\n\x08\x41\x66\x66\x65\x63ted\x12\x1d\n\x07package\x18\x01 \x01(\x0b\x32\x0c.osv.Package\x12\x1a\n\x06ranges\x18\x02 \x03(\x0b\x32\n.osv.Range\x12\x10\n\x08versions\x18\x03 \x03(\t\x12\x33\n\x12\x65\x63osystem_specific\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x32\n\x11\x64\x61tabase_specific\x18\x05 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x1f\n\x08severity\x18\x06 \x03(\x0b\x32\r.osv.Severity\"\x87\x01\n\x08Severity\x12 \n\x04type\x18\x01 \x01(\x0e\x32\x12.osv.Severity.Type\x12\r\n\x05score\x18\x02 \x01(\t\"J\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0b\n\x07\x43VSS_V3\x10\x01\x12\x0b\n\x07\x43VSS_V2\x10\x02\x12\x0b\n\x07\x43VSS_V4\x10\x03\x12\n\n\x06Ubuntu\x10\x04\"\x8a\x02\n\x06\x43redit\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontact\x18\x02 \x03(\t\x12\x1e\n\x04type\x18\x03 \x01(\x0e\x32\x10.osv.Credit.Type\"\xc0\x01\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\t\n\x05OTHER\x10\x01\x12\n\n\x06\x46INDER\x10\x02\x12\x0c\n\x08REPORTER\x10\x03\x12\x0b\n\x07\x41NALYST\x10\x04\x12\x0f\n\x0b\x43OORDINATOR\x10\x05\x12\x19\n\x15REMEDIATION_DEVELOPER\x10\x06\x12\x18\n\x14REMEDIATION_REVIEWER\x10\x07\x12\x18\n\x14REMEDIATION_VERIFIER\x10\x08\x12\x08\n\x04TOOL\x10\t\x12\x0b\n\x07SPONSOR\x10\n\"\xd1\x01\n\tReference\x12!\n\x04type\x18\x01 \x01(\x0e\x32\x13.osv.Reference.Type\x12\x0b\n\x03url\x18\x02 \x01(\t\"\x93\x01\n\x04Type\x12\x08\n\x04NONE\x10\x00\x12\x0c\n\x08\x41\x44VISORY\x10\x02\x12\x0b\n\x07\x41RTICLE\x10\x06\x12\r\n\tDETECTION\x10\t\x12\x0e\n\nDISCUSSION\x10\x08\x12\x0c\n\x08\x45VIDENCE\x10\x07\x12\x07\n\x03\x46IX\x10\x04\x12\x0e\n\nINTRODUCED\x10\n\x12\x0b\n\x07PACKAGE\x10\x05\x12\n\n\x06REPORT\x10\x03\x12\x07\n\x03WEB\x10\x01\"\xcd\x03\n\rVulnerability\x12\x16\n\x0eschema_version\x18\x12 \x01(\t\x12\n\n\x02id\x18\x01 \x01(\t\x12-\n\tpublished\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08modified\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twithdrawn\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x61liases\x18\x08 \x03(\t\x12\x0f\n\x07related\x18\r \x03(\t\x12\x10\n\x08upstream\x18\x0e \x03(\t\x12\x0f\n\x07summary\x18\x03 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x04 \x01(\t\x12\x1f\n\x08\x61\x66\x66\x65\x63ted\x18\x11 \x03(\x0b\x32\r.osv.Affected\x12\"\n\nreferences\x18\x10 \x03(\x0b\x32\x0e.osv.Reference\x12\x32\n\x11\x64\x61tabase_specific\x18\x0f \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x1f\n\x08severity\x18\x13 \x03(\x0b\x32\r.osv.Severity\x12\x1c\n\x07\x63redits\x18\x14 \x03(\x0b\x32\x0b.osv.CreditB2Z0github.com/ossf/osv-schema/bindings/go/osvschemab\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'vulnerability_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z0github.com/ossf/osv-schema/bindings/go/osvschema'
  _globals['_COMMIT']._serialized_start=91
  _globals['_COMMIT']._serialized_end=212
  _globals['_COMMIT_REPOTYPE']._serialized_start=176
  _globals['_COMMIT_REPOTYPE']._serialized_end=212
  _globals['_PACKAGE']._serialized_start=214
  _globals['_PACKAGE']._serialized_end=270
  _globals['_EVENT']._serialized_start=272
  _globals['_EVENT']._serialized_end=352
  _globals['_RANGE']._serialized_start=355
  _globals['_RANGE']._serialized_end=496
  _globals['_RANGE_TYPE']._serialized_start=437
  _globals['_RANGE_TYPE']._serialized_end=496
  _globals['_AFFECTED']._serialized_start=499
  _globals['_AFFECTED']._serialized_end=724
  _globals['_SEVERITY']._serialized_start=727
  _globals['_SEVERITY']._serialized_end=862
  _globals['_SEVERITY_TYPE']._serialized_start=788
  _globals['_SEVERITY_TYPE']._serialized_end=862
  _globals['_CREDIT']._serialized_start=865
  _globals['_CREDIT']._serialized_end=1131
  _globals['_CREDIT_TYPE']._serialized_start=939
  _globals['_CREDIT_TYPE']._serialized_end=1131
  _globals['_REFERENCE']._serialized_start=1134
  _globals['_REFERENCE']._serialized_end=1343
  _globals['_REFERENCE_TYPE']._serialized_start=1196
  _globals['_REFERENCE_TYPE']._serialized_end=1343
  _globals['_VULNERABILITY']._serialized_start=1346
  _globals['_VULNERABILITY']._serialized_end=1807
# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Datastore types."""

import datetime
import enum
import logging
import re
import os

from urllib.parse import urlparse
from typing import Self

from google.cloud import ndb
from google.protobuf import json_format, timestamp_pb2
from osv import importfinding_pb2

# pylint: disable=relative-beyond-top-level
from . import bug
from . import ecosystems
from . import gcs
from . import pubsub
from . import purl_helpers
from . import semver_index
from . import sources
from . import utils
from . import vulnerability_pb2

SCHEMA_VERSION = '1.7.3'

_MAX_GIT_VERSIONS_TO_INDEX = 5000

_EVENT_ORDER = {
    'introduced': 0,
    'last_affected': 1,
    'fixed': 2,
    'limit': 3,
}


def _check_valid_severity(prop, value):
  """Check valid severity."""
  del prop

  if value not in ('NEGLIGIBLE', 'LOW', 'MEDIUM', 'HIGH', 'CRITICAL'):
    raise ValueError('Invalid severity: ' + value)


def _check_valid_range_type(prop, value):
  """Check valid range type."""
  del prop

  if value not in ('GIT', 'SEMVER', 'ECOSYSTEM'):
    raise ValueError('Invalid range type: ' + value)


def _check_valid_event_type(prop, value):
  """Check valid event type."""
  del prop

  if value not in ('introduced', 'fixed', 'last_affected', 'limit'):
    raise ValueError('Invalid event type: ' + value)


def utcnow():
  """For mocking."""
  return datetime.datetime.now(datetime.UTC)


def _get_purl_indexes(affected_packages):
  """Get list of purls from affected packages, with and without qualifiers"""
  resulting_set = set()
  for pkg in affected_packages:
    if pkg.package.purl:
      resulting_set.add(pkg.package.purl)
      if '?' in pkg.package.purl:
        resulting_set.add(pkg.package.purl.split('?')[0])
  return list(resulting_set)


def _repo_name(repo_url: str) -> str:
  # https://github.com/eclipse-openj9/openj9 -> openj9
  url = urlparse(repo_url)
  assumed_reponame = os.path.dirname(url.path).lstrip("/")
  name = assumed_reponame.rstrip(".git")
  return name


def maybe_strip_repo_prefixes(versions: list[str], repo_urls: list[str]) -> str:
  """Try to strip the repo name from tags prior to normalizing.

  There are some particularly regex-unfriendly tag names that prefix the
  reponame that end in a number, like "openj9-0.8.0", resulting in an
  incorrectly normalized version.
  """

  repo_stripped_versions = versions

  for repo_url in repo_urls:
    assumed_reponame = _repo_name(repo_url).lower()
    repo_stripped_versions = [
        v.lstrip(assumed_reponame).lstrip("-") for v in versions
    ]

  return repo_stripped_versions


# --- OSS-Fuzz-related Entities ---


class IDCounter(ndb.Model):
  """Counter for ID allocations."""
  # Next ID to allocate.
  next_id: int = ndb.IntegerProperty()


class RegressResult(ndb.Model):
  """Regression results."""
  # The commit hash.
  commit: str = ndb.StringProperty(default='')
  # Vulnerability summary.
  summary: str = ndb.TextProperty()
  # Vulnerability details.
  details: str = ndb.TextProperty()
  # Error (if any).
  error: str = ndb.StringProperty()
  # OSS-Fuzz issue ID.
  issue_id: str = ndb.StringProperty()
  # Project for the bug.
  project: str = ndb.StringProperty()
  # Package ecosystem for the project.
  ecosystem: str = ndb.StringProperty()
  # Repo URL.
  repo_url: str = ndb.StringProperty()
  # Severity of the bug.
  severity: str = ndb.StringProperty(validator=_check_valid_severity)
  # Reference URLs.
  reference_urls: list[str] = ndb.StringProperty(repeated=True)
  # Source timestamp.
  timestamp: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)


class FixResult(ndb.Model):
  """Fix results."""
  # The commit hash.
  commit: str = ndb.StringProperty(default='')
  # Vulnerability summary.
  summary: str = ndb.TextProperty()
  # Vulnerability details.
  details: str = ndb.TextProperty()
  # Error (if any).
  error: str = ndb.StringProperty()
  # OSS-Fuzz issue ID.
  issue_id: str = ndb.StringProperty()
  # Project for the bug.
  project: str = ndb.StringProperty()
  # Package ecosystem for the project.
  ecosystem: str = ndb.StringProperty()
  # Repo URL.
  repo_url: str = ndb.StringProperty()
  # Severity of the bug.
  severity: str = ndb.StringProperty(validator=_check_valid_severity)
  # Reference URLs.
  reference_urls: list[str] = ndb.StringProperty(repeated=True)
  # Source timestamp.
  timestamp: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)


# --- OSV Bug entities ---


class AffectedEvent(ndb.Model):
  """Affected event."""
  type: str = ndb.StringProperty(validator=_check_valid_event_type)
  value: str = ndb.StringProperty()


class AffectedRange2(ndb.Model):
  """Affected range."""
  # Type of range.
  type: str = ndb.StringProperty(validator=_check_valid_range_type)
  # Repo URL.
  repo_url: str = ndb.StringProperty()
  # Events.
  events: list[AffectedEvent] = ndb.LocalStructuredProperty(
      AffectedEvent, repeated=True)


class SourceOfTruth(enum.IntEnum):
  """Source of truth."""
  NONE = 0
  # Internal to OSV (e.g. private OSS-Fuzz bugs).
  INTERNAL = 1
  # Vulnerabilities that are available in a public repo.
  SOURCE_REPO = 2


class Package(ndb.Model):
  """Package."""
  ecosystem: str = ndb.StringProperty()
  name: str = ndb.StringProperty()
  purl: str = ndb.StringProperty()


class Severity(ndb.Model):
  """Severity."""
  type: str = ndb.StringProperty()
  score: str = ndb.StringProperty()


class AffectedPackage(ndb.Model):
  """Affected packages."""
  # The affected package identifier.
  package: Package = ndb.StructuredProperty(Package)
  # The list of affected ranges.
  ranges: list[AffectedRange2] = ndb.LocalStructuredProperty(
      AffectedRange2, repeated=True)
  # The list of explicit affected versions.
  versions: list[str] = ndb.TextProperty(repeated=True)
  # Database specific metadata.
  database_specific: dict = ndb.JsonProperty()
  # Ecosystem specific metadata.
  ecosystem_specific: dict = ndb.JsonProperty()
  # Severity of the bug.
  severities: list[Severity] = ndb.LocalStructuredProperty(
      Severity, repeated=True)


class Credit(ndb.Model):
  """Credits."""
  name: str = ndb.StringProperty()
  contact: list[str] = ndb.StringProperty(repeated=True)
  type: str = ndb.StringProperty()


class Bug(ndb.Model):
  """Bug entity."""
  OSV_ID_PREFIX = 'OSV-'
  # Very large fake version to use when there is no fix available.
  _NOT_FIXED_SEMVER = '999999.999999.999999'

  # Display ID as used by the source database. The full qualified database that
  # OSV tracks this as may be different.
  db_id: str = ndb.StringProperty()
  # Other IDs this bug is known as.
  aliases: list[str] = ndb.StringProperty(repeated=True)
  # Related IDs.
  related: list[str] = ndb.StringProperty(repeated=True)
  # Raw upstream vulnerability IDs from the source - does not include
  # exhaustive transitive upstreams
  upstream_raw: list[str] = ndb.StringProperty(repeated=True)
  # Status of the bug.
  status: int = ndb.IntegerProperty()
  # Timestamp when Bug was allocated.
  timestamp: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # When the entry was last edited.
  last_modified: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # Last modified field of the original imported file
  import_last_modified: datetime.datetime = ndb.DateTimeProperty(
      tzinfo=datetime.UTC)
  # When the entry was withdrawn.
  withdrawn: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # The source identifier.
  # For OSS-Fuzz, this oss-fuzz:<ClusterFuzz testcase ID>.
  # For others this is <source>:<path/to/source>.
  source_id: str = ndb.StringProperty()
  # The main fixed commit (from bisection).
  fixed: str = ndb.StringProperty(default='')
  # The main regressing commit (from bisection).
  regressed: str = ndb.StringProperty(default='')
  # List of affected versions.
  affected: list[str] = ndb.TextProperty(repeated=True)
  # List of normalized versions indexed for fuzzy matching.
  affected_fuzzy: list[str] = ndb.StringProperty(repeated=True)
  # OSS-Fuzz issue ID.
  issue_id: str = ndb.StringProperty()
  # Package URL for this package.
  purl: list[str] = ndb.StringProperty(repeated=True)
  # Project/package name for the bug.
  project: list[str] = ndb.StringProperty(repeated=True)
  # Package ecosystem for the project.
  ecosystem: list[str] = ndb.StringProperty(repeated=True)
  # Summary for the bug.
  summary: str = ndb.TextProperty()
  # Vulnerability details.
  details: str = ndb.TextProperty()
  # Severity of the bug.
  severities: list[Severity] = ndb.LocalStructuredProperty(
      Severity, repeated=True)
  # Credits for the bug.
  credits: list[Credit] = ndb.LocalStructuredProperty(Credit, repeated=True)
  # Whether or not the bug is public (OSS-Fuzz only).
  public: bool = ndb.BooleanProperty()
  # Reference URL types (dict of url -> type).
  reference_url_types: dict = ndb.JsonProperty()
  # Search indices (auto-populated)
  search_indices: list[str] = ndb.StringProperty(repeated=True)
  # Search tags (auto-populated)
  search_tags: list[str] = ndb.StringProperty(repeated=True)
  # Whether or not the bug has any affected versions (auto-populated).
  has_affected: bool = ndb.BooleanProperty()
  # Source of truth for this Bug.
  source_of_truth: SourceOfTruth = ndb.IntegerProperty(
      default=SourceOfTruth.INTERNAL)
  # Whether the bug is fixed (indexed for querying).
  is_fixed: bool = ndb.BooleanProperty()
  # Database specific.
  database_specific: dict = ndb.JsonProperty()
  # Normalized SEMVER fixed indexes for querying.
  semver_fixed_indexes: list[str] = ndb.StringProperty(repeated=True)
  # Affected packages and versions.
  affected_packages: list[AffectedPackage] = ndb.LocalStructuredProperty(
      AffectedPackage, repeated=True)
  # The source of this Bug.
  source: str = ndb.StringProperty()

  def id(self):
    """Get the bug ID."""
    if self.db_id:
      return self.db_id

    # TODO(ochang): Remove once all existing bugs have IDs migrated.
    if re.match(r'^\d+', self.key.id()):
      return self.OSV_ID_PREFIX + self.key.id()

    return self.key.id()

  @property
  def repo_url(self):
    """Repo URL. Assumes there is only ever one repo URL per advisory."""
    for affected_package in self.affected_packages:
      for affected_range in affected_package.ranges:
        if affected_range.repo_url:
          return affected_range.repo_url

    return None

  @classmethod
  def get_by_id(cls, vuln_id, *args, **kwargs) -> Self | None:
    """Overridden get_by_id to handle OSV allocated IDs."""
    result = cls.query(cls.db_id == vuln_id).get()
    if result:
      return result

    # TODO(ochang): Remove once all exsting bugs have IDs migrated.
    if vuln_id.startswith(cls.OSV_ID_PREFIX):
      vuln_id = vuln_id[len(cls.OSV_ID_PREFIX):]

    return super().get_by_id(vuln_id, *args, **kwargs)

  def _pre_put_hook(self):  # pylint: disable=arguments-differ
    """Pre-put hook for populating search indices."""
    search_indices = set()

    search_indices.update(_tokenize(self.id()))

    for pkg in self.affected_packages:
      # Set PURL if it wasn't provided.
      if not pkg.package.purl:
        pkg.package.purl = purl_helpers.package_to_purl(
            ecosystems.normalize(pkg.package.ecosystem), pkg.package.name)

    self.project = list({
        pkg.package.name for pkg in self.affected_packages if pkg.package.name
    })

    if self.repo_url and 'GIT' in self.ecosystem:
      self.project.append(self.repo_url)

    self.project.sort()

    ecosystems_set = {
        pkg.package.ecosystem
        for pkg in self.affected_packages
        if pkg.package.ecosystem
    }

    # Add the Git ecosystem if it has git ranges
    for pkg in self.affected_packages:
      if 'GIT' in ecosystems_set:
        break
      for r in pkg.ranges:
        if r.type == 'GIT':
          ecosystems_set.add('GIT')
          break

    # If a withdrawn record has no affected package,
    # assign an '[EMPTY]' ecosystem value for export.
    if not ecosystems_set:
      ecosystems_set.add('[EMPTY]')

    # For all ecosystems that specify a specific version with colon,
    # also add the base name
    ecosystems_set.update({ecosystems.normalize(x) for x in ecosystems_set})

    # Expand the set to include all ecosystem variants.
    ecosystems_set = ecosystems.add_matching_ecosystems(ecosystems_set)

    self.ecosystem = list(ecosystems_set)
    self.ecosystem.sort()

    self.purl = _get_purl_indexes(self.affected_packages)
    self.purl.sort()

    for project in self.project:
      search_indices.update(_tokenize(project))

    for ecosystem in self.ecosystem:
      search_indices.update(_tokenize(ecosystem))

    for alias in self.aliases:
      search_indices.update(_tokenize(alias))

    # Please note this will not include exhaustive transitive upstream
    # so may not appear for all cases.
    for upstream in self.upstream_raw:
      search_indices.update(_tokenize(upstream))

    for affected_package in self.affected_packages:
      for affected_range in affected_package.ranges:
        if affected_range.repo_url and affected_range.repo_url != '':
          url_no_https = affected_range.repo_url.split('//')[1]  # remove https
          repo_url_indices = url_no_https.split('/')[1:]  # remove domain
          repo_url_indices.append(affected_range.repo_url)  # add full url
          repo_url_indices.append(url_no_https)  # add url without https://
          search_indices.update(repo_url_indices)

    self.search_indices = list(set(search_indices))
    self.search_indices.sort()

    search_tags = set()
    search_tags.add(self.id().lower())
    search_tags.update([p.lower() for p in self.project])
    self.search_tags = list(search_tags)
    self.search_tags.sort()

    self.affected_fuzzy = []
    self.semver_fixed_indexes = []
    self.has_affected = False
    self.is_fixed = False

    for affected_package in self.affected_packages:
      # Indexes used for querying by exact version.
      ecosystem_helper = ecosystems.get(affected_package.package.ecosystem)
      if ecosystem_helper is not None:
        # No need to normalize if the ecosystem is supported.
        self.affected_fuzzy.extend(affected_package.versions)
      else:
        if (not affected_package.package.ecosystem and
            len(affected_package.versions) > _MAX_GIT_VERSIONS_TO_INDEX):
          # Assume that if there is no ecosystem specified, then these versions
          # were enumerated from Git.
          #
          # Mitigate cases where the Git repo tag matching results in too many
          # versions to index for Datastore.
          # It's OK to do this because the primary intended matching mechanism
          # for Git is via commit hash matching instead.
          logging.info(
              'Skipping indexing of git versions for %s '
              'as there are too many (%s).', self.db_id,
              len(affected_package.versions))
        else:
          self.affected_fuzzy.extend(
              bug.normalize_tags(
                  maybe_strip_repo_prefixes(
                      affected_package.versions,
                      [range.repo_url for range in affected_package.ranges])))

      self.has_affected |= bool(affected_package.versions)

      for affected_range in affected_package.ranges:
        fixed_version = None
        for event in affected_range.events:
          # Index used to query by fixed/unfixed.
          if event.type == 'limit':
            self.is_fixed = True
            fixed_version = event.value

          if event.type == 'fixed':
            self.is_fixed = True
            fixed_version = event.value

        if affected_range.type == 'SEMVER':
          # Indexes used for querying by semver.
          fixed = fixed_version or self._NOT_FIXED_SEMVER
          self.semver_fixed_indexes.append(semver_index.normalize(fixed))

        self.has_affected |= (affected_range.type in ('SEMVER', 'ECOSYSTEM'))

    self.affected_fuzzy = list(set(self.affected_fuzzy))
    self.affected_fuzzy.sort()

    if not self.last_modified:
      self.last_modified = utcnow()

    if self.source_id:
      self.source, _ = sources.parse_source_id(self.source_id)

    if not self.source:
      raise ValueError('Source not specified for Bug.')

    if not self.db_id:
      raise ValueError('DB ID not specified for Bug.')

    if not self.key:  # pylint: disable=access-member-before-definition
      source_repo = get_source_repository(self.source)
      if not source_repo:
        raise ValueError(f'{self.db_id} has invalid source {self.source}')

      if source_repo.db_prefix and not any(
          self.db_id.startswith(prefix) for prefix in source_repo.db_prefix):
        raise ValueError(
            f'{self.db_id} has invalid prefix for source {self.source}')

      self.key = ndb.Key(Bug, self.db_id)

    if self.withdrawn:
      self.status = bug.BugStatus.INVALID

  def update_from_vulnerability(self, vulnerability):
    """Set fields from vulnerability. Does not set the ID."""
    self.summary = vulnerability.summary
    self.details = vulnerability.details
    self.reference_url_types = {
        ref.url: vulnerability_pb2.Reference.Type.Name(ref.type)
        for ref in vulnerability.references
    }

    if vulnerability.HasField('modified'):
      self.last_modified = vulnerability.modified.ToDatetime(datetime.UTC)
    if vulnerability.HasField('published'):
      self.timestamp = vulnerability.published.ToDatetime(datetime.UTC)
    if vulnerability.HasField('withdrawn'):
      self.withdrawn = vulnerability.withdrawn.ToDatetime(datetime.UTC)
    else:
      self.withdrawn = None

    self.aliases = list(vulnerability.aliases)
    self.related = list(vulnerability.related)
    self.upstream_raw = list(vulnerability.upstream)

    self.affected_packages = []
    for affected_package in vulnerability.affected:
      current = AffectedPackage()
      current.package = Package(
          name=affected_package.package.name,
          ecosystem=affected_package.package.ecosystem,
          purl=affected_package.package.purl)
      current.ranges = []

      for affected_range in affected_package.ranges:
        current_range = AffectedRange2(
            type=vulnerability_pb2.Range.Type.Name(affected_range.type),
            repo_url=affected_range.repo,
            events=[])

        for evt in affected_range.events:
          if evt.introduced:
            current_range.events.append(
                AffectedEvent(type='introduced', value=evt.introduced))
            continue

          if evt.fixed:
            current_range.events.append(
                AffectedEvent(type='fixed', value=evt.fixed))
            continue

          if evt.last_affected:
            current_range.events.append(
                AffectedEvent(type='last_affected', value=evt.last_affected))
            continue

          if evt.limit:
            current_range.events.append(
                AffectedEvent(type='limit', value=evt.limit))
            continue

        current.ranges.append(current_range)

      current.versions = list(affected_package.versions)
      if affected_package.database_specific:
        current.database_specific = json_format.MessageToDict(
            affected_package.database_specific,
            preserving_proto_field_name=True)

      if affected_package.ecosystem_specific:
        current.ecosystem_specific = json_format.MessageToDict(
            affected_package.ecosystem_specific,
            preserving_proto_field_name=True)

      current.severities = []
      for severity in affected_package.severity:
        current.severities.append(
            Severity(
                type=vulnerability_pb2.Severity.Type.Name(severity.type),
                score=severity.score))

      self.affected_packages.append(current)

    self.severities = []
    for severity in vulnerability.severity:
      self.severities.append(
          Severity(
              type=vulnerability_pb2.Severity.Type.Name(severity.type),
              score=severity.score))

    self.credits = []
    for credit in vulnerability.credits:
      cr = Credit(name=credit.name, contact=list(credit.contact))
      if credit.type:
        cr.type = vulnerability_pb2.Credit.Type.Name(credit.type)
      self.credits.append(cr)

    if vulnerability.database_specific:
      self.database_specific = json_format.MessageToDict(
          vulnerability.database_specific, preserving_proto_field_name=True)

  def to_vulnerability_minimal(self):
    """Convert to Vulnerability proto (minimal)."""
    if self.last_modified:
      modified = timestamp_pb2.Timestamp()
      modified.FromDatetime(self.last_modified)
    else:
      modified = None

    return vulnerability_pb2.Vulnerability(id=self.id(), modified=modified)

  @ndb.tasklet
  def to_vulnerability_minimal_async(self,
                                     include_alias=True,
                                     include_upstream=True):
    """Convert to Vulnerability proto (minimal) asynchronously."""
    modified_times = []
    if self.last_modified:
      modified_times.append(self.last_modified)

    # Fetch the last_modified dates from the upstream/alias groups.
    # TODO(michaelkedar): modified time needs to update if related changes.
    # Probably requires a RelatedGroup entity/cron
    alias_future = get_aliases_async(self.id()) if include_alias else None
    upstream_future = (
        get_upstream_async(self.id()) if include_upstream else None)

    if include_alias:
      alias = yield alias_future
      if alias and alias.last_modified:
        modified_times.append(alias.last_modified)

    if include_upstream:
      upstream = yield upstream_future
      if upstream and upstream.last_modified:
        modified_times.append(upstream.last_modified)

    modified = None
    if modified_times:
      modified = timestamp_pb2.Timestamp()
      modified.FromDatetime(max(modified_times))

    return vulnerability_pb2.Vulnerability(id=self.id(), modified=modified)

  def to_vulnerability(self,
                       include_source=False,
                       include_alias=True,
                       include_upstream=True):
    """Convert to Vulnerability proto."""
    affected = []

    source_link = None
    if self.source and include_source:
      source_repo = get_source_repository(self.source)
      if source_repo and source_repo.link:
        source_link = source_repo.link + sources.source_path(source_repo, self)

    if self.affected_packages:
      for affected_package in self.affected_packages:
        ranges = []
        for affected_range in affected_package.ranges:
          events = []
          for event in affected_range.events:
            kwargs = {event.type: event.value}
            events.append(vulnerability_pb2.Event(**kwargs))

          current_range = vulnerability_pb2.Range(
              type=vulnerability_pb2.Range.Type.Value(affected_range.type),
              repo=affected_range.repo_url,
              events=events)

          ranges.append(current_range)

        current = vulnerability_pb2.Affected(
            package=vulnerability_pb2.Package(
                name=affected_package.package.name,
                ecosystem=affected_package.package.ecosystem,
                purl=affected_package.package.purl),
            ranges=ranges,
            versions=affected_package.versions)

        # Converted CVE records have no package defined.
        # Avoid exporting an empty package field.
        if not current.package.ListFields():
          current.ClearField("package")

        if affected_package.database_specific:
          current.database_specific.update(affected_package.database_specific)

        if source_link:
          current.database_specific.update({'source': source_link})

        if affected_package.ecosystem_specific:
          current.ecosystem_specific.update(affected_package.ecosystem_specific)

        for entry in affected_package.severities:
          current.severity.append(
              vulnerability_pb2.Severity(
                  type=vulnerability_pb2.Severity.Type.Value(entry.type),
                  score=entry.score))

        affected.append(current)

    details = self.details

    # Note that there is further possible mutation of this field below when
    # `include_alias` is True or `include_upstream` is True
    if self.last_modified:
      modified = timestamp_pb2.Timestamp()
      modified.FromDatetime(self.last_modified)
    else:
      modified = None

    if self.withdrawn:
      withdrawn = timestamp_pb2.Timestamp()
      withdrawn.FromDatetime(self.withdrawn)
    else:
      withdrawn = None

    if self.timestamp:
      published = timestamp_pb2.Timestamp()
      published.FromDatetime(self.timestamp)
    else:
      published = None

    references = []
    if self.reference_url_types:
      for url, url_type in self.reference_url_types.items():
        references.append(
            vulnerability_pb2.Reference(
                url=url, type=vulnerability_pb2.Reference.Type.Value(url_type)))

    severity = []
    for entry in self.severities:
      severity.append(
          vulnerability_pb2.Severity(
              type=vulnerability_pb2.Severity.Type.Value(entry.type),
              score=entry.score))

    credits_ = []
    for credit in self.credits:
      cr = vulnerability_pb2.Credit(name=credit.name, contact=credit.contact)
      if credit.type:
        cr.type = vulnerability_pb2.Credit.Type.Value(credit.type)
      credits_.append(cr)

    related = self.related
    aliases = []

    if include_alias:
      related_bugs = Bug.query(
          Bug.related == self.db_id, projection=[Bug.db_id]).fetch()
      related_bug_ids = [bug.db_id for bug in related_bugs]
      related = sorted(list(set(related_bug_ids + self.related)))
      # TODO(michaelkedar): modified time needs to update if related changes.
      # Probably requires a RelatedGroup entity/cron

      alias_group = AliasGroup.query(AliasGroup.bug_ids == self.db_id).get()
      if alias_group:
        aliases = sorted(list(set(alias_group.bug_ids) - {self.db_id}))
        modified = timestamp_pb2.Timestamp()
        modified.FromDatetime(
            max(self.last_modified, alias_group.last_modified))

    upstream = []

    if include_upstream:
      upstream_group = UpstreamGroup.query(
          UpstreamGroup.db_id == self.db_id).get()
      if upstream_group:
        upstream = sorted(upstream_group.upstream_ids)
        modified = timestamp_pb2.Timestamp()
        modified.FromDatetime(
            max(self.last_modified, upstream_group.last_modified))

    result = vulnerability_pb2.Vulnerability(
        schema_version=SCHEMA_VERSION,
        id=self.id(),
        published=published,
        modified=modified,  # Note the three places above where this can be set.
        aliases=aliases,
        related=related,
        upstream=upstream,
        withdrawn=withdrawn,
        summary=self.summary,
        details=details,
        affected=affected,
        severity=severity,
        credits=credits_,
        references=references)

    if self.database_specific:
      result.database_specific.update(self.database_specific)

    return result

  @ndb.tasklet
  def to_vulnerability_async(self,
                             include_source=False,
                             include_alias=False,
                             include_upstream=False):
    """Converts to Vulnerability proto and retrieves aliases asynchronously."""
    # Convert the vulnerability without any subqueries first.
    vulnerability: vulnerability_pb2.Vulnerability = self.to_vulnerability(
        include_source=False, include_alias=False, include_upstream=False)

    # Asynchronously make all necessary subqueries.
    if not self.source:
      include_source = False
    source_future = (
        SourceRepository.get_by_id_async(self.source)
        if include_source else None)
    related_future = (
        get_related_async(vulnerability.id) if include_alias else None)
    alias_future = (
        get_aliases_async(vulnerability.id) if include_alias else None)
    upstream_future = (
        get_upstream_async(vulnerability.id) if include_upstream else None)

    if include_source:
      source_repo = yield source_future
      if source_repo and source_repo.link:
        source_link = source_repo.link + sources.source_path(source_repo, self)
        for affected in vulnerability.affected:
          affected.database_specific.update({'source': source_link})

    if include_alias:
      related_bug_ids = yield related_future
      vulnerability.related[:] = sorted(
          list(set(related_bug_ids + list(vulnerability.related))))
      # TODO(michaelkedar): modified time needs to update if related changes.
      # Probably requires a RelatedGroup entity/cron
      alias_group = yield alias_future
      if alias_group:
        alias_ids = sorted(list(set(alias_group.bug_ids) - {vulnerability.id}))
        vulnerability.aliases[:] = alias_ids
        modified_time = vulnerability.modified.ToDatetime(datetime.UTC)
        modified_time = max(alias_group.last_modified, modified_time)
        vulnerability.modified.FromDatetime(modified_time)

    if include_upstream:
      upstream_group = yield upstream_future
      if upstream_group:
        vulnerability.upstream[:] = upstream_group.upstream_ids
        modified_time = vulnerability.modified.ToDatetime(datetime.UTC)
        modified_time = max(upstream_group.last_modified, modified_time)
        vulnerability.modified.FromDatetime(modified_time)
    return vulnerability

  def _post_put_hook(self: Self, future: ndb.Future):  # pylint: disable=arguments-differ
    """Post-put hook for writing new entities for database migration."""
    # TODO(michaelkedar): Currently, only want to run this on the test instance
    # (or when running tests). Remove this check when we're ready for prod.
    project = utils.get_google_cloud_project()
    if not project:
      logging.error('failed to get GCP project')
    if project not in ('oss-vdb-test', 'test-osv'):
      return
    if future.exception():
      logging.error("Not writing new entities for %s since Bug.put() failed",
                    self.db_id)
      return
    populate_entities_from_bug(self)


def _tokenize(value):
  """Tokenize value for indexing."""
  if not value:
    return []

  value_lower = value.lower()

  # Deconstructs the id given into parts by retrieving parts that are
  # alphanumeric.
  # This addresses special cases like SUSE that include ':' in their id suffix
  tokens = {token for token in re.split(r'\W+', value_lower) if token}
  tokens.add(value_lower)

  # Add subsection combinations from id (split at '-') in the search indices
  # Specifically addresses situation in which UBUNTU-CVE-XXXs don't show up
  # when searching for the CVE-XXX.
  # e.g. `a-b-c-d' becomes ['a-b', 'b-c', 'c-d', 'a-b-c', 'b-c-d', 'a-b-c-d']
  # Does not account for combinations with the suffix sections ':' like SUSE
  parts = value_lower.split('-')
  num_parts = len(parts)
  for length in range(2, num_parts + 1):
    for i in range(num_parts - length + 1):
      sub_parts = parts[i:i + length]
      combo = '-'.join(sub_parts)
      tokens.add(combo)
  return tokens


# --- Vulnerability Entity ---


class Vulnerability(ndb.Model):
  """A Vulnerability entry.
  
  Contains a minimal amount of information of an OSV record, including the
  overall modified date, an some raw fields that are  overwritten by our
  enrichment.
  
  The entity's key/id is ID in OSV.
  """

  # The source identifier.
  # For OSS-Fuzz, this oss-fuzz:<ClusterFuzz testcase ID>.
  # For others this is <source>:<path/to/source>.
  source_id: str = ndb.StringProperty()
  # When this record was truly last modified (including e.g. aliases/upstream).
  modified: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # Whether this record has been withdrawn
  # TODO(michaelkedar): I don't think this is necessary
  is_withdrawn: bool = ndb.BooleanProperty()

  # Raw fields from the original source.
  # The reported modified date in the record.
  modified_raw: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # The reported aliased IDs.
  alias_raw: list[str] = ndb.StringProperty(repeated=True)
  # The reported related IDs.
  related_raw: list[str] = ndb.StringProperty(repeated=True)
  # The reported upstream IDs.
  upstream_raw: list[str] = ndb.StringProperty(repeated=True)


# --- Affected versions for matching ---


class AffectedCommits(ndb.Model):
  """AffectedCommits entry."""
  MAX_COMMITS_PER_ENTITY = 10000

  # The main bug ID.
  bug_id: str = ndb.StringProperty()
  # The commit hash.
  commits: list[bytes] = ndb.BlobProperty(repeated=True, indexed=True)
  # Whether or not the bug is public.
  public: bool = ndb.BooleanProperty()
  # The page for this batch of commits.
  page: int = ndb.IntegerProperty(indexed=False)


class AffectedVersions(ndb.Model):
  """AffectedVersions entry, used for finding matching vulnerabilities within
  the OSV API."""
  # The main vulnerability ID.
  vuln_id: str = ndb.StringProperty()
  # The ecosystem of the affected package.
  ecosystem: str = ndb.StringProperty()
  # The name of the affected package.
  name: str = ndb.StringProperty()

  # Only one of the following should be set:
  # The enumerated affected versions.
  versions: list[str] = ndb.TextProperty(repeated=True)
  # The sorted affected events.
  events: list[AffectedEvent] = ndb.LocalStructuredProperty(
      AffectedEvent, repeated=True)

  def sort_key(self):
    """Key function for comparison and deduplication."""
    return (self.vuln_id, self.ecosystem, self.name, tuple(self.versions),
            tuple((e.type, e.value) for e in self.events))


# --- Website search / list entity ---


class ListedVulnerability(ndb.Model):
  """ListedVulnerability entry, used for the website's /list page."""
  # The entity's key/id is ID in OSV

  # The date the vulnerability was published (for sorting & display).
  published: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  # The ecosystems the vulnerability belongs to (for filtering).
  ecosystems: list[str] = ndb.StringProperty(repeated=True)
  # The list of rendered affected packages (for display).
  # e.g. 'PyPI/urllib3', 'github.com/torvalds/linux'
  packages: list[str] = ndb.TextProperty(repeated=True)
  # The summary line (for display).
  summary: str = ndb.TextProperty()
  # Whether there is a fix available (for display).
  is_fixed: bool = ndb.BooleanProperty(indexed=False)
  # The severities of the vulnerability (for display).
  severities: list[Severity] = ndb.LocalStructuredProperty(
      Severity, repeated=True)

  # Strings that the search bar may suggest while typing.
  autocomplete_tags: list[str] = ndb.StringProperty(repeated=True)
  # Strings this matches when searching.
  search_indices: list[str] = ndb.StringProperty(repeated=True)

  @classmethod
  def from_vulnerability(
      cls: Self, vulnerability: vulnerability_pb2.Vulnerability) -> Self:
    """Construct a ListedVulnerability from a complete vulnerability proto"""
    published = vulnerability.published.ToDatetime(datetime.UTC)
    summary = vulnerability.summary
    # TODO(michaelkedar): Take the first line of details if summary is missing.

    all_ecosystems = set()
    all_packages = set()

    is_fixed = False
    severities = set()
    for sev in vulnerability.severity:
      severities.add(
          (vulnerability_pb2.Severity.Type.Name(sev.type), sev.score))

    search_indices = set()
    search_indices.update(_tokenize(vulnerability.id))
    autocomplete_tags = {vulnerability.id.lower()}

    for alias in vulnerability.aliases:
      search_indices.update(_tokenize(alias))
    for upstream in vulnerability.upstream:
      search_indices.update(_tokenize(upstream))
    # related intentionally omitted

    for affected in vulnerability.affected:
      if affected.package.name:
        search_indices.update(_tokenize(affected.package.name))
        autocomplete_tags.add(affected.package.name.lower())
        all_packages.add(affected.package.ecosystem + '/' +
                         affected.package.name)
      if affected.package.ecosystem:
        all_ecosystems.add(affected.package.ecosystem)
      for sev in affected.severity:
        severities.add(
            (vulnerability_pb2.Severity.Type.Name(sev.type), sev.score))
      for r in affected.ranges:
        if r.type == vulnerability_pb2.Range.Type.GIT:
          all_ecosystems.add('GIT')
          search_indices.add(r.repo)
          autocomplete_tags.add(r.repo.lower())
          split = r.repo.split('//')
          if len(split) >= 2:
            no_http = split[1]
            all_packages.add(no_http)
            search_indices.add(no_http)
            # Add the path components exluding the domain name
            search_indices.update(no_http.split('/')[1:])
          else:
            all_packages.add(r.repo)

        if any(e.fixed or e.limit for e in r.events):
          is_fixed = True

    for eco in all_ecosystems:
      # TODO(michaelkedar): Seems like a noisy/useless search index?
      search_indices.update(_tokenize(eco))
      if (e := ecosystems.remove_variants(eco)) is not None:
        search_indices.update(_tokenize(e))

    ecos = sorted({ecosystems.normalize(e) for e in all_ecosystems})
    pkgs = sorted(all_packages)
    sevs = [Severity(type=t, score=s) for t, s in sorted(severities)]
    search_indices = sorted(search_indices)
    autocomplete_tags = sorted(autocomplete_tags)

    return cls(
        id=vulnerability.id,
        published=published,
        ecosystems=ecos,
        packages=pkgs,
        summary=summary,
        is_fixed=is_fixed,
        severities=sevs,
        autocomplete_tags=autocomplete_tags,
        search_indices=search_indices,
    )


def populate_entities_from_bug(entity: Bug):
  """Puts entities (Vulnerability, ListedVulnerability, AffectedVersions) from
  a given Bug entity, and writes completed OSV records to GCS bucket."""
  if not entity.public or entity.status == bug.BugStatus.UNPROCESSED:
    # OSS-Fuzz private Bugs
    return

  vuln_pb = entity.to_vulnerability(
      include_source=True, include_alias=True, include_upstream=True)

  def transaction():
    to_put = []
    to_delete = []
    vuln = Vulnerability.get_by_id(entity.db_id)
    if vuln is None:
      vuln = Vulnerability(id=entity.db_id)
    if vuln.modified != vuln_pb.modified.ToDatetime(datetime.UTC):
      vuln.source_id = entity.source_id
      vuln.modified = vuln_pb.modified.ToDatetime(datetime.UTC)
      vuln.is_withdrawn = entity.withdrawn is not None
      vuln.modified_raw = entity.import_last_modified
      vuln.alias_raw = entity.aliases
      vuln.related_raw = entity.related
      vuln.upstream_raw = entity.upstream_raw
      to_put.append(vuln)

    old_affected = AffectedVersions.query(
        AffectedVersions.vuln_id == entity.db_id).fetch()
    if vuln.is_withdrawn:
      # We do not want the vuln to be searchable if it's been withdrawn.
      to_delete.append(ndb.Key(ListedVulnerability, vuln_pb.id))
      to_delete.extend(av.key for av in old_affected)
    else:
      to_put.append(ListedVulnerability.from_vulnerability(vuln_pb))
      new_affected = affected_from_bug(entity)
      added, removed = diff_affected_versions(old_affected, new_affected)
      to_put.extend(added)
      to_delete.extend(r.key for r in removed)

    ndb.put_multi(to_put)
    ndb.delete_multi(to_delete)

  ndb.transaction(transaction)
  try:
    gcs.upload_vulnerability(vuln_pb)
  except Exception:
    # Writing to bucket failed for some reason. Send a pub/sub message to retry.
    logging.error('Writing to bucket failed for %s', entity.db_id)
    data = vuln_pb.SerializeToString(deterministic=True)
    pubsub.publish_failure(data, type='gcs_retry')


def affected_from_bug(entity: Bug) -> list[AffectedVersions]:
  """Compute the AffectedVersions from a Bug entity."""
  affected_versions = []
  for affected in entity.affected_packages:
    pkg_ecosystem = affected.package.ecosystem
    # Make sure we capture all possible ecosystem variants for matching.
    # e.g. {'Ubuntu:22.04:LTS', 'Ubuntu:22.04', 'Ubuntu'}
    all_pkg_ecosystems = {pkg_ecosystem, ecosystems.normalize(pkg_ecosystem)}
    if (e := ecosystems.remove_variants(pkg_ecosystem)) is not None:
      all_pkg_ecosystems.add(e)

    pkg_name = ecosystems.maybe_normalize_package_names(affected.package.name,
                                                        pkg_ecosystem)

    # Ecosystem helper for sorting the events.
    e_helper = ecosystems.get(pkg_ecosystem)
    # TODO(michaelkedar): I am matching the current behaviour of the API,
    # where GIT tags match to the first git repo in the ranges list, even if
    # there are non-git ranges or multiple git repos in a range.
    repo_url = ''
    for r in affected.ranges:
      if r.type == 'GIT':
        if not repo_url:
          repo_url = r.repo_url
        continue
      if r.type not in ('SEMVER', 'ECOSYSTEM'):
        logging.warning('Unknown range type "%s" in %s', r.type, entity.db_id)
        continue

      events = r.events
      if e_helper is not None:
        # If we have an ecosystem helper sort the events to help with querying.
        events.sort(key=lambda e, sort_key=e_helper.sort_key:
                    (sort_key(e.value), _EVENT_ORDER.get(e.type, -1)))
      # If we don't have an ecosystem helper, assume the events are in order.
      for e in all_pkg_ecosystems:
        affected_versions.append(
            AffectedVersions(
                vuln_id=entity.db_id,
                ecosystem=e,
                name=pkg_name,
                events=events,
            ))

    # Add the enumerated versions
    if affected.versions:
      if pkg_name:  # We need at least a package name to perform matching.
        for e in all_pkg_ecosystems:
          affected_versions.append(
              AffectedVersions(
                  vuln_id=entity.db_id,
                  ecosystem=e,
                  name=pkg_name,
                  versions=affected.versions,
              ))
      if repo_url:
        affected_versions.append(
            AffectedVersions(
                vuln_id=entity.db_id,
                ecosystem='GIT',
                name=normalize_repo_package(repo_url),
                versions=affected.versions,
            ))

  # Deduplicate and sort the affected_versions
  unique_affected_dict = {av.sort_key(): av for av in affected_versions}
  affected_versions = sorted(
      unique_affected_dict.values(), key=AffectedVersions.sort_key)

  return affected_versions


def diff_affected_versions(
    old: list[AffectedVersions], new: list[AffectedVersions]
) -> tuple[list[AffectedVersions], list[AffectedVersions]]:
  """Find all the AffectedVersion entities that were added/removed from `old` to
  get `new` (ignoring the entity IDs).
  
  returns (added, removed)
  """
  all_dict = {av.sort_key(): av for av in old + new}
  old_set = {av.sort_key() for av in old}
  new_set = {av.sort_key() for av in new}

  added_keys = new_set - old_set
  removed_keys = old_set - new_set

  added = [all_dict[k] for k in added_keys]
  removed = [all_dict[k] for k in removed_keys]

  return added, removed


def normalize_repo_package(repo_url: str) -> str:
  """Normalize the repo_url for use with GIT AffectedVersions entities.
  
  Removes the scheme/protocol and the .git extension, and trailing slashes.
  
  For example:
    - 'http://git.musl-libc.org/git/musl' (e.g. CVE-2017-15650)
      and 'https://git.musl-libc.org/git/musl' (e.g. CVE-2025-26519)
      both become 'git.musl-libc.org/git/musl'
    - 'https://github.com/curl/curl.git' (e.g. CURL-CVE-2024-2004)
      and 'https://github.com/curl/curl' (e.g. CVE-2025-5025)
      both become 'github.com/curl/curl'
  """
  if not repo_url:
    return repo_url

  try:
    parsed = urlparse(repo_url)
    # Remove scheme and reconstruct without it
    # Keep netloc (hostname) and path
    normalized = parsed.netloc + parsed.path

    # Remove trailing slash
    normalized = normalized.rstrip('/')
    normalized = normalized.removesuffix('.git')
    return normalized
  except Exception:
    return repo_url


# --- Indexer entities ---


class RepoIndex(ndb.Model):
  """RepoIndex entry"""
  # The dependency name
  name: str = ndb.StringProperty()
  # The base cpe without the version
  base_cpe: str = ndb.StringProperty()
  # The repository commit
  commit: bytes = ndb.BlobProperty()
  # The source address
  repo_addr: str = ndb.StringProperty()
  # The scanned file extensions
  file_exts: list[str] = ndb.StringProperty(repeated=True)
  # The hash algorithm used
  file_hash_type: str = ndb.StringProperty()
  # The repository type
  repo_type: str = ndb.StringProperty()
  # A bitmap of which buckets are empty
  empty_bucket_bitmap: bytes = ndb.BlobProperty()
  # Number of files in this repo
  file_count: int = ndb.IntegerProperty()
  # Tag name of the source
  tag: str = ndb.StringProperty()


class FileResult(ndb.Model):
  """FileResult entry containing the path and hash"""
  # The hash value of the file
  hash: bytes = ndb.BlobProperty(indexed=True)
  # The file path
  path: str = ndb.TextProperty()


class RepoIndexBucket(ndb.Model):
  """RepoIndexResult entries containing the actual hash values"""
  # The file results per file
  node_hash: bytes = ndb.BlobProperty(indexed=True)
  # number of files this hash represents
  files_contained: int = ndb.IntegerProperty()


# --- SourceRepository ---


class SourceRepositoryType(enum.IntEnum):
  """SourceRepository type."""
  GIT = 0
  BUCKET = 1
  REST_ENDPOINT = 2


class SourceRepository(ndb.Model):
  """Source repository."""
  # The SourceRepositoryType of the repository.
  type: int = ndb.IntegerProperty()
  # The name of the source.
  name: str = ndb.StringProperty()
  # The repo URL for the source for SourceRepositoryType.GIT.
  repo_url: str = ndb.StringProperty()
  # The username to use for SSH auth for SourceRepositoryType.GIT.
  repo_username: str = ndb.StringProperty()
  # Optional branch for repo for SourceRepositoryType.GIT.
  repo_branch: str = ndb.StringProperty()
  # The API endpoint for SourceRepositoryType.REST_ENDPOINT.
  rest_api_url: str = ndb.StringProperty()
  # Bucket name for SourceRepositoryType.BUCKET.
  bucket: str = ndb.StringProperty()
  # Vulnerability data not under this path is ignored by the importer.
  directory_path: str = ndb.StringProperty()
  # Last synced hash for SourceRepositoryType.GIT.
  last_synced_hash: str = ndb.StringProperty()
  # Last date recurring updates were requested.
  last_update_date: datetime.datetime = ndb.DateTimeProperty(
      tzinfo=datetime.UTC)
  # Patterns of files to exclude (regex).
  ignore_patterns: list[str] = ndb.StringProperty(repeated=True)
  # Whether this repository is editable.
  editable: bool = ndb.BooleanProperty(default=False)
  # Default extension.
  extension: str = ndb.StringProperty(default='.yaml')
  # Key path within each file to store the vulnerability.
  key_path: str = ndb.StringProperty()
  # If true, don't analyze any Git ranges.
  ignore_git: bool = ndb.BooleanProperty(default=False)
  # Whether to detect cherypicks or not (slow for large repos).
  detect_cherrypicks: bool = ndb.BooleanProperty(default=True)
  # Whether to consider all branches when analyzing GIT ranges.
  consider_all_branches: bool = ndb.BooleanProperty(default=False)
  # Whether to populate "affected[].versions" from Git ranges.
  versions_from_repo: bool = ndb.BooleanProperty(default=True)
  # Ignore last import time once (SourceRepositoryType.BUCKET).
  ignore_last_import_time: bool = ndb.BooleanProperty(default=False)
  # HTTP link prefix to individual OSV source records.
  link: str = ndb.StringProperty()
  # HTTP link prefix to individual vulnerability records for humans.
  human_link: str = ndb.StringProperty()
  # DB prefix, if the database allocates its own.
  # https://ossf.github.io/osv-schema/#id-modified-fields
  db_prefix: list[str] = ndb.StringProperty(repeated=True)
  # Apply strict validation (JSON Schema + linter checks) to this source.
  strict_validation: bool = ndb.BooleanProperty(default=False)

  def ignore_file(self, file_path):
    """Return whether or not we should be ignoring a file."""
    if not self.ignore_patterns:
      return False

    file_name = os.path.basename(file_path)
    for pattern in self.ignore_patterns:
      if re.match(pattern, file_name):
        return True

    return False

  def _pre_put_hook(self):  # pylint: disable=arguments-differ
    """Pre-put hook for validation."""
    if self.type == SourceRepositoryType.BUCKET and self.editable:
      raise ValueError('BUCKET SourceRepository cannot be editable.')


# --- Alias & Upstream ---


class AliasGroup(ndb.Model):
  """Alias group."""
  bug_ids: list[str] = ndb.StringProperty(repeated=True)
  last_modified: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)


class AliasAllowListEntry(ndb.Model):
  """Alias group allow list entry."""
  bug_id: str = ndb.StringProperty()


class AliasDenyListEntry(ndb.Model):
  """Alias group deny list entry."""
  bug_id: str = ndb.StringProperty()


class UpstreamGroup(ndb.Model):
  """Upstream group for storing transitive upstreams of a Bug
     This group is in a separate table in order to prevent additional race
     conditions. This makes sure that only the worker is modifying the Bug
     entity directly. 
  """
  # Database ID of the corresponding Bug
  db_id: str = ndb.StringProperty()
  # List of transitive upstreams
  upstream_ids: list[str] = ndb.StringProperty(repeated=True)
  upstream_hierarchy: str = ndb.JsonProperty()
  # Date when group was last modified
  last_modified: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)


# --- ImportFinding ---
# TODO(gongh@): redesign this to make it easy to scale.
class ImportFindings(enum.IntEnum):
  """The possible quality findings about an individual record."""
  UNKNOWN = -1
  NONE = 0
  DELETED = 1
  INVALID_JSON = 2
  INVALID_PACKAGE = 3
  INVALID_PURL = 4
  INVALID_VERSION = 5
  INVALID_COMMIT = 6
  INVALID_RANGE = 7
  INVALID_RECORD = 8
  INVALID_ALIASES = 9
  INVALID_UPSTREAM = 10
  INVALID_RELATED = 11
  BAD_ALIASED_CVE = 12


class ImportFinding(ndb.Model):
  """Quality findings about an individual record."""
  bug_id: str = ndb.StringProperty()
  source: str = ndb.StringProperty()
  findings: list[ImportFindings] = ndb.IntegerProperty(repeated=True)
  first_seen: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)
  last_attempt: datetime.datetime = ndb.DateTimeProperty(tzinfo=datetime.UTC)

  def _pre_put_hook(self):  # pylint: disable=arguments-differ
    """Pre-put hook for setting key."""
    if not self.key:  # pylint: disable=access-member-before-definition
      self.key = ndb.Key(ImportFinding, self.bug_id)

  def to_proto(self):
    """Converts to ImportFinding proto."""
    return importfinding_pb2.ImportFinding(
        bug_id=self.bug_id,
        source=self.source,
        findings=self.findings,  # type: ignore
        first_seen=self.first_seen.timestamp_pb(),  #type: ignore
        last_attempt=self.last_attempt.timestamp_pb(),  #type: ignore
    )


def get_source_repository(source_name: str) -> SourceRepository:
  """Get source repository."""
  return SourceRepository.get_by_id(source_name)


def sorted_events(ecosystem, range_type, events) -> list[AffectedEvent]:
  """Sort events."""
  if range_type == 'GIT':
    # No need to sort.
    return events

  if range_type == 'SEMVER':
    ecosystem_helper = ecosystems.SemverEcosystem()
  else:
    ecosystem_helper = ecosystems.get(ecosystem)

  if ecosystem_helper is None:
    raise ValueError('Unsupported ecosystem ' + ecosystem)

  # Remove any magic '0' values.
  sorted_copy = []
  zero_event = None
  for event in events:
    if event.value == '0':
      zero_event = event
      continue

    sorted_copy.append(event)

  sorted_copy.sort(key=lambda e: ecosystem_helper.sort_key(e.value))
  if zero_event:
    sorted_copy.insert(0, zero_event)

  return sorted_copy


@ndb.tasklet
def get_aliases_async(bug_id: str) -> ndb.Future:
  """Gets aliases asynchronously."""
  alias_group = yield AliasGroup.query(AliasGroup.bug_ids == bug_id).get_async()
  return alias_group


@ndb.tasklet
def get_related_async(bug_id: str) -> ndb.Future:
  """Gets related bugs asynchronously."""
  related_bugs = yield Bug.query(
      Bug.related == bug_id, projection=[Bug.db_id]).fetch_async()
  related_bug_ids = [bug.db_id for bug in related_bugs]
  return related_bug_ids


@ndb.tasklet
def get_upstream_async(bug_id: str) -> ndb.Future:
  """Gets upstream bugs asynchronously."""
  upstream_group = yield UpstreamGroup.query(
      UpstreamGroup.db_id == bug_id).get_async()
  return upstream_group



#!/usr/bin/env python3
"""
Generates Go constants from enums in the OSV JSON schema by dynamically
discovering enum definitions.

This script works by recursively scanning the JSON schema for properties that
are of type 'string' and have an 'enum' field. For each enum it finds, it
generates a corresponding Go type and a const block containing all the enum's
possible values.

The naming of the generated Go types is determined by heuristics based on the
enum's location (path) within the schema.
"""

import json
import os

# The version is not in the schema, so it's hardcoded here for now.
# This should be kept in sync with the canonical version.
SCHEMA_VERSION = "1.7.3"


def to_pascal_case(s: str) -> str:
    """Converts a snake_case, kebab-case, or space-separated string to PascalCase."""
    s = s.replace('-', '_').replace(' ', '_')
    return ''.join(word.capitalize() for word in s.split('_'))


def convert_enum_value_to_go_name(type_name: str, value: str) -> str:
    """
    Converts an enum value string (e.g., "CVSS_V2") into its Go constant
    name suffix (e.g., "CvssV2").
    """
    if type_name == 'Ecosystem':
        # Handle special cases for ecosystems that don't fit the standard
        # PascalCase conversion (e.g., 'crates.io').
        if value == 'crates.io':
            return 'CratesIO'
        if value == 'npm':
            return 'NPM'
        # General case for ecosystems.
        name = value[0].upper() + value[1:]
        name = name.replace('-', '').replace(' ', '')
        return name

    # For all other enum types, convert the value to PascalCase.
    # e.g., "CVSS_V2" -> "CvssV2"
    # e.g., "REMEDIATION_DEVELOPER" -> "RemediationDeveloper"
    return to_pascal_case(value)


def generate_go_enum(type_name: str, enum_values: list[str]) -> str:
    """
    Generates a Go type definition and a corresponding const block for a
    given enum.
    """
    # Determine the prefix for the constant names by removing 'Type' from the end
    # of the type name. For example, 'CreditType' -> 'Credit'.
    prefix = type_name.replace('Type', '')

    lines = []
    lines.append(f'type {type_name} string')
    lines.append('')
    lines.append('const (')

    constants = []
    # Sort values for deterministic output and handle potential duplicates from
    # different parts of the schema being merged.
    for value in sorted(list(set(enum_values))):
        # Construct the full Go constant name, e.g., "Credit" + "Reporter".
        go_name = prefix + convert_enum_value_to_go_name(type_name, value)
        constants.append((go_name, value))

    # Sort by the generated Go constant name case-insensitively.
    constants.sort(key=lambda x: x[0].casefold())

    if not constants:
        lines.append(')')
        return '\n'.join(lines)

    # Find the longest constant name to align the values for better readability.
    longest_name = max(len(c[0]) for c in constants)

    for name, value in constants:
        lines.append(f'\t{name.ljust(longest_name)} {type_name} = "{value}"')

    lines.append(')')
    return '\n'.join(lines)


def get_type_name_from_path(path: tuple) -> str:
    """
    Derives a descriptive Go type name from a path in the JSON schema.

    This function contains heuristics to generate a good type name. For example,
    if an enum is found at the path ('properties', 'credits', 'items', 'properties', 'type'),
    it will identify 'credits' and 'type' and return 'CreditType'.

    For newly defined enums not covered by the special cases, it falls back to
    using the property name where the enum is defined.
    """
    # Look for specific, well-known paths first to give them clear,
    # explicit names.
    if 'credits' in path and 'type' in path:
        return 'CreditType'
    if 'references' in path and 'type' in path:
        return 'ReferenceType'
    if 'severity' in path and 'type' in path:
        return 'SeverityType'
    if 'ranges' in path and 'type' in path:
        return 'RangeType'
    if 'ecosystemName' in path:
        return 'Ecosystem'

    # Fallback for less specific paths or new enums.
    if path:
        # The property name is usually the last element in the path.
        name_key = str(path[-1])
        # If the enum is defined directly under a 'type' property, the actual
        # name of the field is likely a few levels up the path.
        if name_key == 'type' and len(path) > 2:
            name_key = str(path[-3])
        return to_pascal_case(name_key) + 'Type'

    return 'UnknownType'


def find_enums(schema, path=(), enums=None):
    """
    Recursively traverses the JSON schema to find all string-based enums.
    """
    if enums is None:
        enums = {}

    if isinstance(schema, dict):
        # An enum is identified by the presence of an 'enum' key with a list
        # value, and a 'type' key with the value 'string'.
        if 'enum' in schema and isinstance(
                schema['enum'], list) and schema.get('type') == 'string':

            # Derive a type name (e.g., 'CreditType') from the enum's path.
            type_name = get_type_name_from_path(path)

            # Group all enum values by their derived type name. This handles
            # cases where the same enum is defined in multiple places.
            if type_name not in enums:
                enums[type_name] = []
            enums[type_name].extend(schema['enum'])

        # Continue the recursive search down the schema tree.
        for key, value in schema.items():
            find_enums(value, path + (key, ), enums)

    elif isinstance(schema, list):
        # If the current schema element is a list, iterate through its items.
        for i, item in enumerate(schema):
            find_enums(item, path + (i, ), enums)

    return enums


def main():
    """Main function to generate the Go constants file."""
    # Construct absolute paths to the input schema and the output file.
    script_dir = os.path.dirname(os.path.abspath(__file__))
    schema_path = os.path.join(script_dir, '..', 'validation', 'schema.json')
    output_path = os.path.join(script_dir, '..', 'bindings', 'go', 'osvschema',
                               'constants.go')

    try:
        with open(schema_path, 'r', encoding='utf-8') as f:
            schema = json.load(f)
    except FileNotFoundError:
        print(f"Error: '{schema_path}' not found.")
        return

    # Scan the schema to find all enums.
    enums = find_enums(schema)

    # Prepare the content for the generated Go file.
    output_parts = [
        '// Code generated by scripts/generate-go-constants.py. DO NOT EDIT.',
        'package osvschema', '', f'const SchemaVersion = "{SCHEMA_VERSION}"',
        ''
    ]

    # Generate the Go code for each discovered enum, in alphabetical order.
    for type_name in sorted(enums.keys()):
        enum_values = enums[type_name]
        output_parts.append(generate_go_enum(type_name, enum_values))
        output_parts.append('')

    # Write the generated content to the output file.
    output_content = '\n'.join(output_parts)
    with open(output_path, 'w', encoding='utf-8') as f:
        f.write(output_content)

    print(f"Generated Go constants in {output_path}")


if __name__ == '__main__':
    main()



"""Test Intermediate OSV object creation"""
import unittest

from redhat_osv.csaf import CSAF
from redhat_osv.osv import OSV, Event


class ScoreTest(unittest.TestCase):
    """Tests OSV vulnerability scores"""

    test_csaf_files = [
        "rhsa-2003_315.json", "rhsa-2015_0008.json"
    ]

    def test_missing_cvss_v3(self):
        """Test parsing a CSAF file with missing CVSSv3 score"""
        for test_csaf_file in self.test_csaf_files:
            csaf_file = f"testdata/CSAF/{test_csaf_file}"
            with open(csaf_file, "r", encoding="utf-8") as fp:
                csaf_data = fp.read()
            csaf = CSAF(csaf_data)
            with self.subTest(csaf_file):
                assert csaf
                assert len(csaf.vulnerabilities) == 1
                assert not csaf.vulnerabilities[0].cvss_v3_base_score
                for vuln in csaf.vulnerabilities:
                    for remediation in vuln.remediations:
                        assert "@" in remediation.purl

                # See https://github.com/ossf/osv-schema/pull/308#issuecomment-2456061864
                osv = OSV(csaf, "test_date")
                assert not hasattr(osv, "severity")
                for affected in osv.affected:
                    assert "@" not in affected.package.purl


class EventTest(unittest.TestCase):
    """ Tests OSV affected range events"""

    def test_init_event(self):
        """Test parsing various Events"""
        event = Event("introduced")
        assert event.event_type == "introduced"
        assert event.version == "0"

        event = Event("fixed", "1")
        assert event.event_type == "fixed"
        assert event.version == "1"

        with self.assertRaises(ValueError):
            Event("test")


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 OSV Schema Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GHSA to JSON dumper."""
import argparse
import json
import os

import requests

_BASE_QUERY = """
{
  securityAdvisories(first: 100 %(query)s %(cursor)s) {
    edges { cursor
      node {
        ghsaId
        identifiers {
          value
        }
        references {
          url
        }
        description
        summary
        severity
        cvss {
          score
          vectorString
        }
        cwes(first: 32) {
          nodes {
            cweId
            description
            name
          }
        }
        permalink
        publishedAt
        updatedAt
        withdrawnAt
        vulnerabilities(first: 32) {
          nodes {
            package {
              ecosystem
              name
            }
            firstPatchedVersion {
              identifier
            }
            vulnerableVersionRange
          }
        }
      }
    }
    pageInfo {
      hasNextPage
    }
  }
}
"""


def run_graphql(query: str, token: str):
    """Runs a GraphQL query."""
    response = requests.post(
        'https://api.github.com/graphql',
        json={'query': query},
        headers={'Authorization': 'Bearer ' + token})
    response.raise_for_status()
    return response.json()


def dump(out_dir: str, token: str, query: str):
    """Dumps advisories."""
    count = 0
    cursor_arg = ''

    while True:
        result = run_graphql(_BASE_QUERY % {'cursor':cursor_arg, 'query':query}, token)

        if 'data' not in result:
            print('Got invalid response', result)
            raise Exception('Invalid response')

        for edge in result['data']['securityAdvisories']['edges']:
            node = edge['node']
            with open(os.path.join(out_dir, node['ghsaId'] + '.json'),
                      'w') as handle:
                handle.write(json.dumps(node))

            count += 1
            if count % 500 == 0:
                print(f'Up to {count} advisories.')

            cursor = edge['cursor']
            cursor_arg = f'after: "{cursor}"'

        if not result['data']['securityAdvisories']['pageInfo'].get(
                'hasNextPage'):
            break

    print(f'Dumped {count} advisories.')


def main():
    """Main function."""
    parser = argparse.ArgumentParser(description='GHSA dumper.')
    parser.add_argument('--token', help='GitHub API token', required=True)
    parser.add_argument('--query', help='GitHub Security Advisory Query')
    parser.add_argument('out_dir', help='Output directory')

    args = parser.parse_args()
    dump(args.out_dir, args.token, args.query)


if __name__ == '__main__':
    main()
